import logging
from typing import List, Set

from peewee import fn, SqliteDatabase

from cozy.db.book import Book as BookModel
from cozy.db.file import File
from cozy.db.track import Track
from cozy.db.track_to_file import TrackToFile
from cozy.ext import inject
from cozy.extensions.is_same_book import is_same_book
from cozy.media.media_file import MediaFile
from cozy.model.book import Book, BookIsEmpty

log = logging.getLogger("db_importer")


class TrackInsertRequest:
    track_data: object
    file: File
    start_at: int

    def __init__(self, track_data: object, file: File, start_at: int):
        self.track_data = track_data
        self.file = file
        self.start_at = start_at


class BookUpdatePositionRequest:
    book_id: int
    progress: int

    def __init__(self, book_id: int, progress: int):
        self.book_id = book_id
        self.progress = progress


class DatabaseImporter:
    _db = inject.attr(SqliteDatabase)

    def __init__(self):
        self._book_update_positions: List[BookUpdatePositionRequest] = []

    def insert_many(self, media_files: Set[MediaFile]):
        self._book_update_positions = []

        files = self._prepare_files_db_objects(media_files)
        File.insert_many(files).execute()
        tracks = self._prepare_track_db_objects(media_files)
        self._insert_tracks(tracks)
        self._update_book_positions()

    def _prepare_files_db_objects(self, media_files: Set[MediaFile]) -> List[object]:
        files = []

        for media_file in media_files:
            query = File.select().where(File.path == media_file.path)

            if query.exists():
                self._update_files_in_db(query.get(), media_file)
                continue

            file_already_in_list = any(f["path"] == media_file.path for f in files)

            if not file_already_in_list:
                files.append({"path": media_file.path, "modified": media_file.modified})

        return files

    def _update_files_in_db(self, file: File, media_file: MediaFile):
        file.modified = media_file.modified
        file.save(only=file.dirty_fields)

    def _prepare_track_db_objects(self, media_files: Set[MediaFile]) -> Set[TrackInsertRequest]:
        book_db_objects: Set[BookModel] = set()

        for media_file in media_files:
            if not media_file:
                continue

            book = next((book for book in book_db_objects if is_same_book(book.name, media_file.book_name)), None)
            file_query = File.select().where(File.path == media_file.path)
            if not file_query.exists():
                log.error("No file object with path present: {}".format(media_file.path))
                continue

            file = file_query.get()

            if not book:
                book = self._import_or_update_book(media_file)
                book_db_objects.add(book)

            try:
                book_model = Book(self._db, book)
                progress = book_model.progress
            except BookIsEmpty:
                progress = 0

            self._delete_tracks_from_db(media_file)
            tracks = self._get_track_list_for_db(media_file, book)

            for track in tracks:
                start_at = track.pop("startAt")
                yield TrackInsertRequest(track, file, start_at)

            update_position_request_present = any(b.book_id == book.id for b in self._book_update_positions)
            if progress > 0 and not update_position_request_present:
                self._book_update_positions.append(BookUpdatePositionRequest(book.id, progress))

    def _import_or_update_book(self, media_file):
        if BookModel.select(BookModel.name).where(self._matches_db_book(media_file.book_name)).count() < 1:
            book = self._create_book_db_object(media_file)
        else:
            book = self._update_book_db_object(media_file)
        return book

    def _matches_db_book(self, book_name: str) -> bool:
        return fn.Lower(BookModel.name) == book_name.lower()

    def _get_track_list_for_db(self, media_file: MediaFile, book: BookModel):
        tracks = []

        for chapter in media_file.chapters:
            tracks.append({
                "name": chapter.name,
                "number": chapter.number,
                "disk": media_file.disk,
                "book": book,
                "length": chapter.length,
                "startAt": chapter.position,
                "position": 0
            })

        return tracks

    def _update_book_db_object(self, media_file: MediaFile) -> BookModel:
        BookModel.update(name=media_file.book_name,
                         author=media_file.author,
                         reader=media_file.reader,
                         cover=media_file.cover) \
            .where(self._matches_db_book(media_file.book_name)) \
            .execute()
        return BookModel.select().where(self._matches_db_book(media_file.book_name)).get()

    def _create_book_db_object(self, media_file: MediaFile) -> BookModel:
        return BookModel.create(name=media_file.book_name,
                                author=media_file.author,
                                reader=media_file.reader,
                                cover=media_file.cover,
                                position=0,
                                rating=-1)

    def _get_track_db_objects_for_media_file(self, media_file: MediaFile) -> List[Track]:
        all_track_mappings = TrackToFile.select().join(File).where(TrackToFile.file.path == media_file.path)

        for item in all_track_mappings:
            yield item.track

    def _delete_tracks_from_db(self, media_file: MediaFile):
        for track in self._get_track_db_objects_for_media_file(media_file):
            track.delete_instance(recursive=True)

    def _is_chapter_count_in_db_different(self, media_file: MediaFile) -> bool:
        all_track_mappings = self._get_chapter_count_in_db(media_file)

        if all_track_mappings != len(media_file.chapters):
            return True
        else:
            return False

    def _get_chapter_count_in_db(self, media_file: MediaFile) -> int:
        all_track_mappings = TrackToFile.select().join(File).where(TrackToFile.file.path == media_file.path)

        return all_track_mappings.count()

    def _insert_tracks(self, tracks: Set[TrackInsertRequest]):
        for track in tracks:
            track_db = Track.insert(track.track_data).execute()
            TrackToFile.create(track=track_db, file=track.file, start_at=track.start_at)

    def _update_book_positions(self):
        for book_position in self._book_update_positions:
            book = BookModel.get_or_none(book_position.book_id)

            if not book:
                log.error("Could not restore book position because book is not present")
                continue

            self._update_book_position(book, book_position.progress)

        self._book_update_positions = []

    def _update_book_position(self, book: BookModel, progress: int):
        try:
            book_model = Book(self._db, book)
        except BookIsEmpty:
            log.error("Could not restore book position because book is empty")
            return

        completed_chapter_length = 0
        for chapter in book_model.chapters:
            old_position = progress
            if completed_chapter_length + chapter.length > old_position:
                chapter.position = chapter.start_position + ((old_position - completed_chapter_length) * 10 ** 9)
                book_model.position = chapter.id
                return
            else:
                completed_chapter_length += chapter.length

        book_model.position = 0



import time


def timing(f):
    def wrap(*args):
        time1 = time.time()
        ret = f(*args)
        time2 = time.time()
        print('{:s} function took {:.3f} ms'.format(f.__name__, (time2-time1)*1000.0))

        return ret
    return wrap


import traceback

from cozy.report.log_level import LogLevel

from multiprocessing.pool import ThreadPool as Pool

from cozy.report.report_to_loki import report

report_pool = Pool(5)

def info(component: str, message: str):
    report_pool.apply_async(report, [component, LogLevel.INFO, message, None])


def warning(component: str, message: str):
    report_pool.apply_async(report, [component, LogLevel.WARNING, message, None])


def error(component: str, message: str):
    report_pool.apply_async(report, [component, LogLevel.ERROR, message, None])


def exception(component: str, exception: Exception, message=None):
    if not message:
        message = traceback.format_exc()

    report_pool.apply_async(report, [component, LogLevel.ERROR, message, exception])


def close():
    report_pool.close()
    report_pool.terminate()



from gi.repository import Gtk, Pango, GObject, Gdk

from cozy.control.string_representation import seconds_to_str
from cozy.model.chapter import Chapter


@Gtk.Template.from_resource('/com/github/geigi/cozy/chapter_element.ui')
class ChapterElement(Gtk.EventBox):
    __gtype_name__ = "ChapterElement"

    icon_event_box: Gtk.EventBox = Gtk.Template.Child()
    icon_stack: Gtk.Stack = Gtk.Template.Child()
    number_label: Gtk.Label = Gtk.Template.Child()
    play_icon: Gtk.Image = Gtk.Template.Child()
    title_label: Gtk.Label = Gtk.Template.Child()
    duration_label: Gtk.Label = Gtk.Template.Child()

    def __init__(self, chapter: Chapter):
        self.selected = False
        self.chapter: Chapter = chapter

        super().__init__()

        self.connect("enter-notify-event", self._on_enter_notify)
        self.connect("leave-notify-event", self._on_leave_notify)
        self.connect("button-press-event", self._on_button_press)
        self.set_tooltip_text(_("Play this part"))

        # This box contains all content
        self.box = Gtk.Box()
        self.box.set_orientation(Gtk.Orientation.HORIZONTAL)
        self.box.set_spacing(3)
        self.box.set_halign(Gtk.Align.FILL)
        self.box.set_valign(Gtk.Align.CENTER)

        # These are the widgets that contain data
        self.play_img = Gtk.Image()
        no_label = Gtk.Label()
        title_label = Gtk.Label()
        dur_label = Gtk.Label()

        self.play_img.set_margin_right(5)
        self.play_img.props.width_request = 16

        if self.chapter.number > 0:
            no_label.set_text(str(self.chapter.number))
        no_label.props.margin = 4
        no_label.set_margin_right(7)
        no_label.set_margin_left(0)
        no_label.set_size_request(30, -1)
        no_label.set_xalign(1)

        self.number_label.set_text(str(self.chapter.number))
        self.title_label.set_text(self.chapter.name)
        self.duration_label.set_text(seconds_to_str(self.chapter.length))

        self.icon_event_box.connect("enter-notify-event", self._on_enter_notify)
        self.icon_event_box.connect("leave-notify-event", self._on_leave_notify)

    def _on_button_press(self, _, event):
        if event.type == Gdk.EventType.BUTTON_PRESS and event.button != 1:
            return False

        self.emit("play-pause-clicked", self.chapter)
        return True

    def _on_enter_notify(self, _, __):
        if not self.selected:
            self.icon_stack.set_visible_child_name("play_icon")
        self.get_style_context().add_class("box_hover")
        self.play_icon.get_style_context().add_class("box_hover")

    def _on_leave_notify(self, _, __):
        self.get_style_context().remove_class("box_hover")
        self.play_icon.get_style_context().remove_class("box_hover")

        if not self.selected:
            self.icon_stack.set_visible_child_name("number")

    def select(self):
        self.selected = True
        self.icon_stack.set_visible_child_name("play_icon")

    def deselect(self):
        self.selected = False
        self.icon_stack.set_visible_child_name("number")

    def set_playing(self, playing):
        if playing:
            self.play_icon.set_from_icon_name("pause-symbolic", Gtk.IconSize.SMALL_TOOLBAR)
        else:
            self.play_icon.set_from_icon_name("play-symbolic", Gtk.IconSize.SMALL_TOOLBAR)


GObject.type_register(ChapterElement)
GObject.signal_new('play-pause-clicked', ChapterElement, GObject.SIGNAL_RUN_LAST, GObject.TYPE_PYOBJECT,
                   (GObject.TYPE_PYOBJECT,))



from typing import List
from packaging import version

import gi

from cozy.application_settings import ApplicationSettings
from cozy.ext import inject
from cozy.ui.main_view import CozyUI
from cozy.version import __version__ as CozyVersion

gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, Handy


@Gtk.Template(resource_path='/com/github/geigi/cozy/whats_new.ui')
class WhatsNewWindow(Handy.Window):
    __gtype_name__ = 'WhatsNew'

    content_stack: Gtk.Stack = Gtk.Template.Child()
    continue_button: Gtk.Button = Gtk.Template.Child()
    children: List[Gtk.Widget]

    main_window: CozyUI = inject.attr("MainWindow")
    app_settings: ApplicationSettings = inject.attr(ApplicationSettings)

    page = 0

    def __init__(self, **kwargs):
        if self.app_settings.last_launched_version == CozyVersion:
            return

        super().__init__(**kwargs)

        self.set_modal(self.main_window.window)

        self._fill_window()
        if len(self.children) < 1:
            self.end()
            return

        self.set_default_size(800, 550)

        for widget in self.children:
            self.content_stack.add(widget)
            widget.set_visible(False)

        self.children[0].set_visible(True)
        self.continue_button.connect("clicked", self.__on_continue_clicked)
        self.show()

    def _fill_window(self):
        self.children = []

        last_launched_version = version.parse(self.app_settings.last_launched_version)

        if type(last_launched_version) is version.LegacyVersion:
            self._fill_welcome()
        else:
            self._fill_whats_new(last_launched_version)

    def _fill_welcome(self):
        from cozy.ui.widgets.welcome import Welcome
        from cozy.ui.widgets.error_reporting import ErrorReporting
        self.children.append(Welcome())
        self.children.append(ErrorReporting())

    def _fill_whats_new(self, last_launched_version: version.Version):
        from cozy.ui.widgets.whats_new_m4b_chapter import INTRODUCED
        if last_launched_version < version.parse(INTRODUCED):
            from cozy.ui.widgets.whats_new_m4b_chapter import WhatsNewM4BChapter
            self.children.append(WhatsNewM4BChapter())

        from cozy.ui.widgets.whats_new_library import INTRODUCED
        if last_launched_version < version.parse(INTRODUCED):
            from cozy.ui.widgets.whats_new_library import WhatsNewLibrary
            self.children.append(WhatsNewLibrary())

        from cozy.ui.widgets.whats_new_m4b import INTRODUCED
        if last_launched_version < version.parse(INTRODUCED):
            from cozy.ui.widgets.whats_new_m4b import WhatsNewM4B
            self.children.append(WhatsNewM4B())

    def __on_continue_clicked(self, widget):
        if len(self.children) == self.page + 1:
            self.end()
            return

        self.children[self.page].set_visible(False)
        self.page += 1
        self.content_stack.set_visible_child(self.children[self.page])
        self.children[self.page].set_visible(True)

    def end(self):
        self.app_settings.last_launched_version = CozyVersion
        self.close()
        self.destroy()



import threading
from threading import Thread

from cozy.ext import inject
from cozy.ui.widgets.search_results import BookSearchResult, ArtistSearchResult

import gi

from cozy.view_model.search_view_model import SearchViewModel

gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, GLib


# TODO: There is a lot of app logic in this class that should be in the view model.
#  Ideally this class only retrieves lists of results from the view model and only handles displaying them.
class SearchView:
    view_model = inject.attr(SearchViewModel)

    search_thread = None
    search_thread_stop = None

    def __init__(self):
        self.builder = Gtk.Builder.new_from_resource("/com/github/geigi/cozy/search_popover.ui")

        self.popover = self.builder.get_object("search_popover")

        self.book_label = self.builder.get_object("book_label")
        self.track_label = self.builder.get_object("track_label")
        self.author_label = self.builder.get_object("author_label")
        self.reader_label = self.builder.get_object("reader_label")
        self.reader_box = self.builder.get_object("reader_result_box")
        self.author_box = self.builder.get_object("author_result_box")
        self.book_box = self.builder.get_object("book_result_box")
        self.track_box = self.builder.get_object("track_result_box")
        self.entry = self.builder.get_object("search_entry")
        self.scroller = self.builder.get_object("search_scroller")
        self.book_separator = self.builder.get_object("book_separator")
        self.author_separator = self.builder.get_object("author_separator")
        self.reader_separator = self.builder.get_object("reader_separator")
        self.stack = self.builder.get_object("search_stack")

        self.entry.connect("search-changed", self.__on_search_changed)

        if Gtk.get_minor_version() > 20:
            self.scroller.set_max_content_width(400)
            self.scroller.set_max_content_height(600)
            self.scroller.set_propagate_natural_height(True)
            self.scroller.set_propagate_natural_width(True)

        self.search_thread = Thread(target=self.search, name="SearchThread")
        self.search_thread_stop = threading.Event()

        self._connect_view_model()

    def _connect_view_model(self):
        self.view_model.bind_to("search_open", self._on_search_open_changed)

    def search(self, user_search: str):
        # we need the main context to call methods in the main thread after the search is finished
        main_context = GLib.MainContext.default()

        books = list({
            book
            for book
            in self.view_model.books
            if user_search.lower() in book.name.lower()
               or user_search.lower() in book.author.lower()
               or user_search.lower() in book.reader.lower()
        })
        books = sorted(books, key=lambda book: book.name.lower())
        if self.search_thread_stop.is_set():
            return
        main_context.invoke_full(
            GLib.PRIORITY_DEFAULT, self.__on_book_search_finished, books)

        authors = sorted({
            author
            for author
            in self.view_model.authors
            if user_search.lower() in author.lower()
        })
        if self.search_thread_stop.is_set():
            return
        main_context.invoke_full(
            GLib.PRIORITY_DEFAULT, self.__on_author_search_finished, authors)

        readers = sorted({
            reader
            for reader
            in self.view_model.readers
            if user_search.lower() in reader.lower()
        })
        if self.search_thread_stop.is_set():
            return
        main_context.invoke_full(
            GLib.PRIORITY_DEFAULT, self.__on_reader_search_finished, readers)

        if len(readers) < 1 and len(authors) < 1 and len(books) < 1:
            main_context.invoke_full(
                GLib.PRIORITY_DEFAULT, self.stack.set_visible_child_name, "nothing")

    def close(self, object=None):
        if Gtk.get_minor_version() < 22:
            self.popover.hide()
        else:
            self.popover.popdown()

    def __on_search_changed(self, sender):
        self.search_thread_stop.set()

        # we want to avoid flickering of the search box size
        # as we remove widgets and add them again
        # so we get the current search popup size and set it as
        # the preferred size until the search is finished
        # this helps only a bit, the widgets are still flickering
        self.popover.set_size_request(self.popover.get_allocated_width(),
                                      self.popover.get_allocated_height())

        # hide nothing found
        self.stack.set_visible_child_name("main")

        # First clear the boxes
        self.book_box.remove_all_children()
        self.author_box.remove_all_children()
        self.reader_box.remove_all_children()

        # Hide all the labels & separators
        self.book_label.set_visible(False)
        self.author_label.set_visible(False)
        self.reader_label.set_visible(False)
        self.book_separator.set_visible(False)
        self.author_separator.set_visible(False)
        self.reader_separator.set_visible(False)
        self.track_label.set_visible(False)

        user_search = self.entry.get_text()
        if user_search:
            if self.search_thread.is_alive():
                self.search_thread.join(timeout=0.2)
            self.search_thread_stop.clear()
            self.search_thread = Thread(
                target=self.search, args=(user_search,))
            self.search_thread.start()
        else:
            self.stack.set_visible_child_name("start")
            self.popover.set_size_request(-1, -1)

    def _on_search_open_changed(self):
        if self.view_model.search_open == False:
            self.close()

    def __on_book_search_finished(self, books):
        if len(books) > 0:
            self.stack.set_visible_child_name("main")
            self.book_label.set_visible(True)
            self.book_separator.set_visible(True)

            for book in books:
                if self.search_thread_stop.is_set():
                    return

                book_result = BookSearchResult(book, self.view_model.jump_to_book)
                self.book_box.add(book_result)

    def __on_author_search_finished(self, authors):
        if len(authors) > 0:
            self.stack.set_visible_child_name("main")
            self.author_label.set_visible(True)
            self.author_separator.set_visible(True)

            for author in authors:
                if self.search_thread_stop.is_set():
                    return

                author_result = ArtistSearchResult(self.view_model.jump_to_author, author, True)
                self.author_box.add(author_result)

    def __on_reader_search_finished(self, readers):
        if len(readers) > 0:
            self.stack.set_visible_child_name("main")
            self.reader_label.set_visible(True)
            self.reader_separator.set_visible(True)

            for reader in readers:
                if self.search_thread_stop.is_set():
                    return

                reader_result = ArtistSearchResult(self.view_model.jump_to_reader, reader, False)
                self.reader_box.add(reader_result)

        self.popover.set_size_request(-1, -1)



import logging

import gi

from cozy.ui.media_controller_big import MediaControllerBig
from cozy.ui.media_controller_small import MediaControllerSmall

gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, Handy

log = logging.getLogger("MediaController")


class MediaController:
    def __init__(self, main_window_builder: Gtk.Builder):
        super().__init__()

        self._media_control_squeezer: Handy.Squeezer = main_window_builder.get_object("media_control_squeezer")
        self._media_controller_small: MediaControllerSmall = MediaControllerSmall()
        self._media_controller_big: MediaControllerBig = MediaControllerBig()
        self._media_control_squeezer.add(self._media_controller_big)
        self._media_control_squeezer.add(self._media_controller_small)



class Singleton(type):
    _instances = {}
    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]
        


import gi

from cozy.ext import inject
from cozy.view_model.playback_speed_view_model import PlaybackSpeedViewModel

gi.require_version('Gtk', '3.0')
from gi.repository import Gtk


@Gtk.Template.from_resource('/com/github/geigi/cozy/playback_speed_popover.ui')
class PlaybackSpeedPopover(Gtk.Popover):
    __gtype_name__ = "PlaybackSpeedPopover"

    _view_model: PlaybackSpeedViewModel = inject.attr(PlaybackSpeedViewModel)

    playback_speed_scale: Gtk.Scale = Gtk.Template.Child()
    playback_speed_label: Gtk.Label = Gtk.Template.Child()

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.playback_speed_scale.add_mark(1.0, Gtk.PositionType.RIGHT, None)
        self.playback_speed_scale.set_increments(0.02, 0.05)
        self.playback_speed_scale.connect("value-changed", self._on_playback_speed_scale_changed)

        self._connect_view_model()
        self._on_playback_speed_changed()

    def _connect_view_model(self):
        self._view_model.bind_to("playback_speed", self._on_playback_speed_changed)

    def _on_playback_speed_scale_changed(self, _):
        speed = round(self.playback_speed_scale.get_value(), 2)
        self._view_model.playback_speed = speed

        self.playback_speed_label.set_markup("<span font_features='tnum'>{speed:3.1f} x</span>".format(speed=speed))

    def _on_playback_speed_changed(self):
        self.playback_speed_scale.set_value(self._view_model.playback_speed)



import os

from gi.repository import Gtk

import cozy.ui
from cozy.ext import inject
from cozy.media.importer import Importer
from cozy.model.chapter import Chapter
from cozy.model.library import Library


class FileNotFoundDialog():
    _importer: Importer = inject.attr(Importer)
    _library: Library = inject.attr(Library)


    def __init__(self, chapter: Chapter):
        self.missing_chapter = chapter
        self.parent = cozy.ui.main_view.CozyUI()
        self.builder = Gtk.Builder.new_from_resource(
            "/com/github/geigi/cozy/file_not_found.ui")
        self.dialog = self.builder.get_object("dialog")
        self.dialog.set_modal(self.parent.window)
        self.builder.get_object("file_label").set_markup(
            "<tt>" + chapter.file + "</tt>")

        cancel_button = self.builder.get_object("cancel_button")
        cancel_button.connect("clicked", self.close)
        locate_button = self.builder.get_object("locate_button")
        locate_button.connect("clicked", self.locate)

    def show(self):
        self.dialog.show()

    def close(self, _):
        self.dialog.destroy()

    def locate(self, __):
        directory, filename = os.path.split(self.missing_chapter.file)
        dialog = Gtk.FileChooserDialog("Please locate the file " + filename, self.parent.window,
                                       Gtk.FileChooserAction.OPEN,
                                       (Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
                                        Gtk.STOCK_OPEN, Gtk.ResponseType.OK))

        filter = Gtk.FileFilter()
        filter.add_pattern(filename)
        filter.set_name(filename)
        dialog.add_filter(filter)
        path, file_extension = os.path.splitext(self.missing_chapter.file)
        filter = Gtk.FileFilter()
        filter.add_pattern("*" + file_extension)
        filter.set_name(file_extension + " files")
        dialog.add_filter(filter)
        filter = Gtk.FileFilter()
        filter.add_pattern("*")
        filter.set_name(_("All files"))
        dialog.add_filter(filter)
        dialog.set_local_only(False)

        response = dialog.run()
        if response == Gtk.ResponseType.OK:
            new_location = dialog.get_filename()
            self.missing_chapter.file = new_location
            self._importer.scan()
            self.dialog.destroy()

        dialog.destroy()



from enum import Enum, auto


class View(Enum):
    EMPTY_STATE = auto()
    PREPARING_LIBRARY = auto()
    LIBRARY = auto()
    LIBRARY_FILTER = auto()
    LIBRARY_BOOKS = auto()
    BOOK_DETAIL = auto()
    NO_MEDIA = auto()


import gi

gi.require_version('Granite', '1.0')

from gi.repository import Gtk, Granite

from cozy.ext import inject


class InfoBanner:
    _builder: Gtk.Builder = inject.attr("MainWindowBuilder")

    def __init__(self):
        super(InfoBanner, self).__init__()

        self._overlay: Gtk.Overlay = self._builder.get_object("banner_overlay")
        self._toast: Granite.WidgetsToast = Granite.WidgetsToast.new("")
        self._toast.show_all()
        self._overlay.add_overlay(self._toast)
        self._banner_displayed: bool = False

    def show(self, message: str):
        self._toast.set_property("title", message)
        self._toast.set_reveal_child(True)



from peewee import CharField, BooleanField, ForeignKeyField, IntegerField

from cozy.db.model_base import ModelBase
from cozy.db.book import Book

DB_VERSION = 9


class Settings(ModelBase):
    path = CharField()
    first_start = BooleanField(default=True)
    last_played_book = ForeignKeyField(Book, null=True)
    version = IntegerField(default=DB_VERSION)



import logging
import os
import threading
import time
from enum import Enum, auto
from typing import Optional

import gi

from cozy.architecture.event_sender import EventSender
from cozy.report import reporter

gi.require_version('Gst', '1.0')
from gi.repository import Gst

log = logging.getLogger("gst_player")


class GstPlayerState(Enum):
    STOPPED = auto()
    PAUSED = auto()
    PLAYING = auto()


class GstPlayer(EventSender):
    def __init__(self):
        super().__init__()

        self._bus: Optional[Gst.Bus] = None
        self._player: Optional[Gst.Bin] = None
        self._bus_signal_id: Optional[int] = None
        self._playback_speed: float = 1.0
        self._playback_speed_timer_running: bool = False
        self._volume: float = 1.0

        Gst.init(None)

    @property
    def position(self) -> int:
        if not self._is_player_loaded():
            return 0

        position = self._query_gst_time(self._player.query_position)

        if position:
            return position

        log.warning("Failed to query position from player.")
        reporter.warning("gst_player", "Failed to query position from player.")

        return 0

    @position.setter
    def position(self, new_position_ns: int):
        new_position_ns = max(0, new_position_ns)
        duration = self._query_gst_time(self._player.query_duration)

        if duration:
            new_position_ns = min(new_position_ns, duration)

        self._execute_seek(new_position_ns)

    @property
    def playback_speed(self) -> float:
        return self._playback_speed

    @playback_speed.setter
    def playback_speed(self, value: float):
        if not self._is_player_loaded():
            return

        self._playback_speed = value

        if self._playback_speed_timer_running:
            return

        self._playback_speed_timer_running = True

        t = threading.Timer(0.2, self._on_playback_speed_timer)
        t.name = "PlaybackSpeedDelayTimer"
        t.start()

    @property
    def loaded_file_path(self) -> Optional[str]:
        if not self._is_player_loaded():
            return None

        uri = self._player.get_property("current-uri")
        if uri:
            return uri.replace("file://", "")
        else:
            return None

    @property
    def state(self) -> GstPlayerState:
        if not self._is_player_loaded():
            return GstPlayerState.STOPPED

        _, state, __ = self._player.get_state(Gst.CLOCK_TIME_NONE)
        if state == Gst.State.PLAYING:
            return GstPlayerState.PLAYING
        elif state == Gst.State.PAUSED:
            return GstPlayerState.PAUSED
        else:
            log.debug("GST player state was not playing or paused but {}.".format(state))
            return GstPlayerState.STOPPED

    @property
    def volume(self) -> float:
        if not self._is_player_loaded():
            log.error("Could not determine volume because player is not loaded.")
            return 1.0

        return self._player.get_property("volume")

    @volume.setter
    def volume(self, new_value: float):
        self._volume = max(0.0, min(1.0, new_value))

        if not self._is_player_loaded():
            log.warning("Could not set volume because player is not loaded.")
            return

        self._player.set_property("volume", self._volume)
        self._player.set_property("mute", False)

    def init(self):
        if self._player:
            self.dispose()

        self._player = Gst.ElementFactory.make("playbin", "player")
        scaletempo = Gst.ElementFactory.make("scaletempo", "scaletempo")
        scaletempo.sync_state_with_parent()

        audiobin = Gst.ElementFactory.make("bin", "audiosink")
        audiobin.add(scaletempo)

        audiosink = Gst.ElementFactory.make("autoaudiosink", "audiosink")
        audiobin.add(audiosink)

        scaletempo.link(audiosink)
        pad = scaletempo.get_static_pad("sink")
        ghost_pad = Gst.GhostPad.new("sink", pad)
        audiobin.add_pad(ghost_pad)

        self._player.set_property("audio-sink", audiobin)

        self._bus = self._player.get_bus()
        self._bus.add_signal_watch()
        self._bus_signal_id = self._bus.connect("message", self._on_gst_message)

    def dispose(self):
        if not self._player:
            return

        if self._bus_signal_id:
            self._bus.disconnect(self._bus_signal_id)

        self._player.set_state(Gst.State.NULL)
        self._playback_speed = 1.0
        log.info("Dispose")
        self.emit_event("dispose")

    def load_file(self, path: str):
        self.init()

        if not os.path.exists(path):
            raise FileNotFoundError()

        self._player.set_property("uri", "file://" + path)
        self._player.set_state(Gst.State.PAUSED)
        self._player.set_property("volume", self._volume)
        self._player.set_property("mute", False)

    def play(self):
        if not self._is_player_loaded() or self.state == GstPlayerState.PLAYING:
            return

        success = self._player.set_state(Gst.State.PLAYING)

        if success == Gst.StateChangeReturn.FAILURE:
            log.warning("Failed set gst player to play.")
            reporter.warning("gst_player", "Failed set gst player to play.")
        else:
            self.emit_event("state", GstPlayerState.PLAYING)

    def pause(self):
        if not self._is_player_loaded():
            return

        success = self._player.set_state(Gst.State.PAUSED)

        if success == Gst.StateChangeReturn.FAILURE:
            log.warning("Failed set gst player to pause.")
            reporter.warning("gst_player", "Failed set gst player to pause.")
        else:
            self.emit_event("state", GstPlayerState.PAUSED)

    def stop(self):
        if not self._is_player_loaded():
            return

        self.dispose()
        self.emit_event("state", GstPlayerState.STOPPED)

    def _is_player_loaded(self) -> bool:
        if not self._player:
            return False

        _, state, __ = self._player.get_state(Gst.CLOCK_TIME_NONE)
        if state != Gst.State.PLAYING and state != Gst.State.PAUSED:
            return False

        return True

    @staticmethod
    def _query_gst_time(query_function) -> Optional[int]:
        success = False
        counter = 0

        while not success and counter < 10:
            success, value = query_function(Gst.Format.TIME)

            if success:
                return value
            else:
                counter += 1
                time.sleep(0.01)

        return None

    def _execute_seek(self, new_position_ns: int):
        counter = 0
        seeked = False
        while not seeked and counter < 500:
            seeked = self._player.seek(self._playback_speed, Gst.Format.TIME, Gst.SeekFlags.FLUSH, Gst.SeekType.SET,
                                       new_position_ns, Gst.SeekType.NONE, 0)

            if not seeked:
                counter += 1
                time.sleep(0.01)
        if not seeked:
            log.info("Failed to seek, counter expired.")
            reporter.warning("gst_player", "Failed to seek, counter expired.")

    def _on_playback_speed_timer(self):
        self._player.seek(self._playback_speed, Gst.Format.TIME, Gst.SeekFlags.FLUSH | Gst.SeekFlags.ACCURATE,
                          Gst.SeekType.SET, self.position, Gst.SeekType.NONE, 0)

        self._playback_speed_timer_running = False

    def _on_gst_message(self, _, message: Gst.Message):
        t = message.type
        if t == Gst.MessageType.BUFFERING:
            if message.percentage < 100:
                self._player.set_state(Gst.State.PAUSED)
                log.info("Buffering…")
            else:
                self._player.set_state(Gst.State.PLAYING)
                log.info("Buffering finished.")
        elif t == Gst.MessageType.EOS:
            self.emit_event("file-finished")
        elif t == Gst.MessageType.ERROR:
            error, debug_msg = message.parse_error()

            if error.code == Gst.ResourceError.NOT_FOUND:
                self.stop()
                self.emit_event("resource-not-found")

                log.warning("gst: Resource not found. Stopping player.")
                reporter.warning("gst_player", "gst: Resource not found. Stopping player.")
                return

            reporter.error("player", "{}: {}".format(error.code, error))
            log.error("{}: {}".format(error.code, error))
            log.debug(debug_msg)
            self.emit_event("error", error)



import logging
from typing import List, Optional

import peewee

import cozy.ext.inject as inject
from peewee import SqliteDatabase

from cozy.db.book import Book
from cozy.db.settings import Settings as SettingsModel
from cozy.db.storage import Storage as StorageModel
from cozy.model.storage import Storage, InvalidPath
from cozy.report import reporter

log = logging.getLogger("model.storage_location")


class Settings:
    _storages: List[Storage] = []
    _db = cache = inject.attr(SqliteDatabase)

    def __init__(self):
        self._db_object: SettingsModel = SettingsModel.get()

    @property
    def first_start(self) -> bool:
        return self._db_object.first_start

    @property
    def last_played_book(self) -> Optional[Book]:
        try:
            return self._db_object.last_played_book
        except peewee.DoesNotExist:
            log.warning("last_played_book references an non existent object. Setting last_played_book to None.")
            reporter.warning("settings_model",
                             "last_played_book references an non existent object. Setting last_played_book to None.")

            self.last_played_book = None
            return None

    @last_played_book.setter
    def last_played_book(self, new_value):
        if new_value:
            self._db_object.last_played_book = new_value._db_object
        else:
            self._db_object.last_played_book = None

        self._db_object.save(only=self._db_object.dirty_fields)

    @property
    def default_location(self):
        return next(location
                    for location
                    in self.storage_locations
                    if location.default)

    @property
    def storage_locations(self):
        if not self._storages:
            self._load_all_storage_locations()

        return self._storages

    @property
    def external_storage_locations(self):
        if not self._storages:
            self._load_all_storage_locations()

        return [storage for storage in self._storages if storage.external]

    def invalidate(self):
        self._storages = []

    def _load_all_storage_locations(self):
        self._storages = []

        for storage_db_obj in StorageModel.select(StorageModel.id):
            try:
                self._storages.append(Storage(self._db, storage_db_obj.id))
            except InvalidPath:
                log.error("Invalid path found in database, skipping: {}".format(storage_db_obj.path))

        self._ensure_default_storage_present()

    def _ensure_default_storage_present(self):
        default_storage_present = any(storage.default
                                      for storage
                                      in self._storages)

        if not default_storage_present and len(self._storages) > 0:
            self._storages[0].default = True



import logging
import os
import time
from threading import Thread
from typing import Optional

from gi.repository import GLib, Gst

from cozy.application_settings import ApplicationSettings
from cozy.architecture.event_sender import EventSender
from cozy.control.offline_cache import OfflineCache
from cozy.ext import inject
from cozy.media.gst_player import GstPlayer, GstPlayerState
from cozy.model.book import Book
from cozy.model.chapter import Chapter
from cozy.model.library import Library
from cozy.report import reporter
from cozy.tools import IntervalTimer
from cozy.ui.file_not_found_dialog import FileNotFoundDialog
from cozy.ui.info_banner import InfoBanner

log = logging.getLogger("mediaplayer")

NS_TO_SEC = 10 ** 9
REWIND_SECONDS = 30


class Player(EventSender):
    _library: Library = inject.attr(Library)
    _app_settings: ApplicationSettings = inject.attr(ApplicationSettings)
    _offline_cache: OfflineCache = inject.attr(OfflineCache)
    _info_bar: InfoBanner = inject.attr(InfoBanner)

    _gst_player: GstPlayer = inject.attr(GstPlayer)

    def __init__(self):
        super().__init__()

        self._book: Optional[Book] = None
        self._play_next_chapter: bool = True

        self._gst_player.add_listener(self._on_gst_player_event)

        self.play_status_updater: IntervalTimer = IntervalTimer(1, self._emit_tick)
        self._fadeout_thread: Optional[Thread] = None

        self._gst_player.init()
        self.volume = self._app_settings.volume

        self._load_last_book()

    def _load_last_book(self):
        last_book = self._library.last_played_book

        if last_book:
            self._continue_book(last_book)
            self._rewind_feature()

    @property
    def loaded_book(self) -> Optional[Book]:
        return self._book

    @property
    def loaded_chapter(self) -> Optional[Chapter]:
        if self._book:
            return self._book.current_chapter
        else:
            return None

    @property
    def playing(self) -> bool:
        return self._gst_player.state == GstPlayerState.PLAYING

    @property
    def position(self) -> int:
        return self._gst_player.position

    @position.setter
    def position(self, new_value: int):
        self._gst_player.position = self.loaded_chapter.start_position + (new_value * NS_TO_SEC)

    @property
    def volume(self) -> float:
        return self._gst_player.volume

    @volume.setter
    def volume(self, new_value: float):
        self._gst_player.volume = new_value
        self._app_settings.volume = new_value

    @property
    def play_next_chapter(self) -> bool:
        return self._play_next_chapter

    @play_next_chapter.setter
    def play_next_chapter(self, value: bool):
        self._play_next_chapter = value

    @property
    def playback_speed(self) -> float:
        return self._gst_player.playback_speed

    @playback_speed.setter
    def playback_speed(self, value: float):
        self._gst_player.playback_speed = value

    def play_pause(self):
        if self._gst_player.state == GstPlayerState.PAUSED:
            self._gst_player.play()
        elif self._gst_player.state == GstPlayerState.PLAYING:
            self._gst_player.pause()
        else:
            log.error("Trying to play/pause although player is in STOP state.")
            reporter.error("player", "Trying to play/pause although player is in STOP state.")

    def pause(self, fadeout: bool = False):
        if fadeout and not self._fadeout_thread:
            log.info("Starting fadeout playback")
            self._fadeout_thread = Thread(target=self._fadeout_playback, name="PlayerFadeoutThread")
            self._fadeout_thread.start()
            return

        if self._gst_player.state == GstPlayerState.PLAYING:
            self._gst_player.pause()

    def play_pause_book(self, book: Book):
        if not book:
            log.error("Cannot play book which is None.")
            reporter.error("player", "Cannot play book which is None.")
            return

        if self._book == book:
            self.play_pause()
        else:
            self._continue_book(book)
            self._gst_player.play()

    def play_pause_chapter(self, book: Book, chapter: Chapter):
        if not book or not chapter:
            log.error("Cannot play chapter which is None.")
            reporter.error("player", "Cannot play chapter which is None.")
            return

        if self._book and self._book.current_chapter == chapter:
            self.play_pause()
            return

        if self._book != book:
            self._load_book(book)

        self._load_chapter(chapter)
        self._gst_player.play()

        book.position = chapter.id

    def rewind(self):
        state = self._gst_player.state
        if state != GstPlayerState.STOPPED:
            self._rewind_in_book()
        if state == GstPlayerState.PLAYING:
            self._gst_player.play()

    def forward(self):
        state = self._gst_player.state
        if state != GstPlayerState.STOPPED:
            self._forward_in_book()
        if state == GstPlayerState.PLAYING:
            self._gst_player.play()

    def destroy(self):
        self._gst_player.dispose()

        self._stop_tick_thread()

        if self._fadeout_thread:
            self._fadeout_thread.stop()

    def _load_book(self, book: Book):
        if self._book == book:
            log.info("Not loading new book because it's unchanged.")
            return

        self._book = book
        self._book.last_played = int(time.time())

    def _continue_book(self, book: Book):
        if self._book == book:
            log.info("Not loading new book because it's unchanged.")
            return

        self._load_book(book)
        self._load_chapter(book.current_chapter)

    def _load_chapter(self, chapter: Chapter):
        file_changed = False

        if not self._book:
            log.error("There is no book loaded but there should be.")
            reporter.error("player", "There is no book loaded but there should be.")
            return

        self._library.last_played_book = self._book
        media_file_path = self._get_playback_path(chapter)

        if self._gst_player.loaded_file_path == media_file_path:
            log.info("Not loading a new file because the new chapter is within the old file.")
        else:
            log.info("Loading new file for chapter.")
            try:
                self._gst_player.load_file(media_file_path)
                file_changed = True
            except FileNotFoundError:
                self._handle_file_not_found()
                return

        if file_changed or self._should_jump_to_chapter_position(chapter.position):
            self._gst_player.position = chapter.position
            self._gst_player.playback_speed = self._book.playback_speed

        if file_changed or self._book.position != chapter.id:
            self._book.position = chapter.id
            self.emit_event_main_thread("chapter-changed", self._book)

    def _get_playback_path(self, chapter: Chapter):
        if self._book.offline and self._book.downloaded:
            path = self._offline_cache.get_cached_path(chapter)

            if path and os.path.exists(path):
                return path

        return chapter.file

    def _rewind_in_book(self):
        if not self._book:
            log.error("Rewind in book not possible because no book is loaded.")
            reporter.error("player", "Rewind in book not possible because no book is loaded.")
            return

        current_position = self._gst_player.position
        current_position_relative = max(current_position - self.loaded_chapter.start_position, 0)
        chapter_number = self._book.chapters.index(self._book.current_chapter)
        rewind_seconds = self._app_settings.rewind_duration * self.playback_speed

        if current_position_relative / NS_TO_SEC - rewind_seconds > 0:
            self._gst_player.position = current_position - NS_TO_SEC * rewind_seconds
        elif chapter_number > 0:
            previous_chapter = self._book.chapters[chapter_number - 1]
            self._load_chapter(previous_chapter)
            self._gst_player.position = previous_chapter.end_position + (
                        current_position_relative - NS_TO_SEC * rewind_seconds)
        else:
            self._gst_player.position = 0

    def _forward_in_book(self):
        if not self._book:
            log.error("Forward in book not possible because no book is loaded.")
            reporter.error("player", "Forward in book not possible because no book is loaded.")
            return

        current_position = self._gst_player.position
        current_position_relative = max(current_position - self.loaded_chapter.start_position, 0)
        old_chapter = self._book.current_chapter
        chapter_number = self._book.chapters.index(self._book.current_chapter)
        forward_seconds = self._app_settings.forward_duration * self.playback_speed

        if current_position_relative / NS_TO_SEC + forward_seconds < self._book.current_chapter.length:
            self._gst_player.position = current_position + (NS_TO_SEC * forward_seconds)
        elif chapter_number < len(self._book.chapters) - 1:
            next_chapter = self._book.chapters[chapter_number + 1]
            self._load_chapter(next_chapter)
            self._gst_player.position = next_chapter.start_position + (
                    NS_TO_SEC * forward_seconds - (old_chapter.length * NS_TO_SEC - current_position_relative))
        else:
            self._next_chapter()

    def _rewind_feature(self):
        if self._app_settings.replay:
            self._rewind_in_book()
            self._emit_tick()

    def _next_chapter(self):
        if not self._book:
            log.error("Cannot play next chapter because no book reference is stored.")
            reporter.error("player", "Cannot play next chapter because no book reference is stored.")

        index_current_chapter = self._book.chapters.index(self._book.current_chapter)

        self._book.current_chapter.position = self._book.current_chapter.start_position
        if len(self._book.chapters) <= index_current_chapter + 1:
            log.info("Book finished, stopping playback.")
            self._finish_book()
            self._gst_player.stop()
        else:
            chapter = self._book.chapters[index_current_chapter + 1]
            chapter.position = chapter.start_position
            self.play_pause_chapter(self._book, chapter)

    def _on_gst_player_event(self, event: str, message):
        if event == "file-finished":
            self._next_chapter()
        elif event == "resource-not-found":
            self._handle_file_not_found()
        elif event == "state" and message == GstPlayerState.PLAYING:
            self._book.last_played = int(time.time())
            self._start_tick_thread()
            self.emit_event_main_thread("play", self._book)
        elif event == "state" and message == GstPlayerState.PAUSED:
            self._stop_tick_thread()
            self.emit_event_main_thread("pause")
        elif event == "state" and message == GstPlayerState.STOPPED:
            self._stop_playback()
        elif event == "error":
            self._handle_gst_error(message)

    def _handle_gst_error(self, error: GLib.Error):
        if error.code != Gst.ResourceError.BUSY:
            self._info_bar.show(error.message)

        if error.code == Gst.ResourceError.OPEN_READ or Gst.ResourceError.READ:
            self._stop_playback()

    def _handle_file_not_found(self):
        if self.loaded_chapter:
            FileNotFoundDialog(self.loaded_chapter).show()
            self._stop_playback()
        else:
            log.warning("No chapter loaded, cannot display file not found dialog.")

    def _stop_playback(self):
        self._stop_tick_thread()
        self._book = None
        self.emit_event_main_thread("pause")
        self.emit_event_main_thread("stop")

    def _finish_book(self):
        if self._book:
            self._book.position = -1
            self._library.last_played_book = None

        self.emit_event_main_thread("book-finished", self._book)

    def _start_tick_thread(self):
        if self.play_status_updater:
            self.play_status_updater.stop()

        self.play_status_updater = IntervalTimer(1, self._emit_tick)
        self.play_status_updater.start()

    def _stop_tick_thread(self):
        if self.play_status_updater:
            self.play_status_updater.stop()
            self.play_status_updater = None

    def _emit_tick(self):
        if not self.loaded_chapter or not self.loaded_book:
            log.info("Not emitting tick because no book/chapter is loaded.")
            return

        if self.position > self.loaded_chapter.end_position:
            self._next_chapter()

        try:
            self.loaded_chapter.position = self.position
            position_for_ui = self.position - self.loaded_chapter.start_position
            self.emit_event_main_thread("position", position_for_ui)
        except Exception as e:
            log.warning("Could not emit position event: {}".format(e))

    def _fadeout_playback(self):
        duration = self._app_settings.sleep_timer_fadeout_duration * 20
        current_vol = self._gst_player.volume
        for i in range(0, duration):
            volume = max(current_vol - (i / duration), 0)
            self._gst_player.position = volume
            time.sleep(0.05)

        log.info("Fadeout completed.")
        self.play_pause()
        self._gst_player.volume = current_vol
        self.emit_event_main_thread("fadeout-finished", None)

        self._fadeout_thread = None

    def _should_jump_to_chapter_position(self, position: int) -> bool:
        """
        Should the player jump to the given position?
        This allows gapless playback for media files that contain many chapters.
        """

        difference = abs(self.position - position)
        if difference < 10 ** 9:
            return False

        return True



from gi.repository import Gtk

import cozy.ui


class ImportFailedDialog():
    """
    Dialog that displays failed files on import.
    """

    def __init__(self, files):
        self.parent = cozy.ui.main_view.CozyUI()
        self.builder = Gtk.Builder.new_from_resource(
            "/com/github/geigi/cozy/import_failed.ui")
        self.dialog = self.builder.get_object("dialog")
        self.dialog.set_modal(self.parent.window)
        self.text = self.builder.get_object("files_buffer")

        files_string = "\n".join(files)
        self.text.set_text(files_string.encode("utf-8", "replace").decode("utf-8"))

        locate_button = self.builder.get_object("ok_button")
        locate_button.connect("clicked", self.ok)

    def show(self):
        """
        show this dialog
        """
        self.dialog.show()

    def ok(self, button):
        """
        Close this dialog and destroy it.
        """
        self.dialog.destroy()



import logging
import os
from enum import Enum, auto
from typing import Optional

from cozy import tools
from cozy.application_settings import ApplicationSettings
from cozy.architecture.observable import Observable
from cozy.ext import inject
from cozy.media.player import Player
from cozy.tools import IntervalTimer

log = logging.getLogger("sleep_timer_view_model")


class SystemPowerControl(Enum):
    OFF = auto()
    SUSPEND = auto()
    SHUTDOWN = auto()


class SleepTimerViewModel(Observable):
    _app_settings: ApplicationSettings = inject.attr(ApplicationSettings)
    _player: Player = inject.attr(Player)

    def __init__(self):
        super().__init__()

        self._remaining_seconds: int = 0
        self._system_power_control: SystemPowerControl = SystemPowerControl.OFF
        self._sleep_timer: Optional[IntervalTimer] = None
        self._wait_for_fadeout_end: bool = False

        self._player.add_listener(self._on_player_changed)

    @property
    def timer_enabled(self) -> bool:
        return self.remaining_seconds > 0 or self.stop_after_chapter

    @property
    def remaining_seconds(self) -> int:
        return self._remaining_seconds

    @remaining_seconds.setter
    def remaining_seconds(self, new_value: int):
        self._remaining_seconds = new_value

        if new_value > 0:
            self._start_timer()
        else:
            self._stop_timer()

        self._notify("timer_enabled")

    @property
    def system_power_control(self) -> SystemPowerControl:
        return self._system_power_control

    @system_power_control.setter
    def system_power_control(self, new_value: SystemPowerControl):
        self._system_power_control = new_value

    @property
    def stop_after_chapter(self) -> bool:
        return not self._player.play_next_chapter

    @stop_after_chapter.setter
    def stop_after_chapter(self, new_value: bool):
        self._player.play_next_chapter = not new_value
        self._stop_timer()
        self.remaining_seconds = 0
        self._notify("remaining_seconds")
        self._notify("timer_enabled")

    def destroy(self):
        self._stop_timer()

    def _start_timer(self):
        if self._sleep_timer:
            return

        if self.remaining_seconds < 1:
            return

        if not self._player.playing:
            return

        log.info("Start Timer")
        self._sleep_timer = tools.IntervalTimer(1, self._on_timer_tick)
        self._sleep_timer.start()

    def _stop_timer(self):
        if not self._sleep_timer:
            return

        log.info("Stop Timer")
        self._sleep_timer.stop()
        self._sleep_timer = None

    def _on_timer_tick(self):
        self.remaining_seconds = self.remaining_seconds - 1
        self._notify_main_thread("remaining_seconds")

        fadeout = self._get_fadeout()
        if self.remaining_seconds - fadeout < 1:
            self._stop_playback()
            self._stop_timer()
            self._notify("timer_enabled")

            if not self._wait_for_fadeout_end:
                self._handle_system_power_event()
            else:
                self.remaining_seconds = 0
                self._notify_main_thread("remaining_seconds")

    def _get_fadeout(self) -> int:
        fadeout = 0

        if self._app_settings.sleep_timer_fadeout:
            fadeout = self._app_settings.sleep_timer_fadeout_duration

        return fadeout

    def _stop_playback(self):
        fadeout = self._get_fadeout()
        self._wait_for_fadeout_end = fadeout > 0
        self._player.pause(fadeout=fadeout > 0)

    def _handle_system_power_event(self):
        platform = tools.system_platform()
        command = ""

        if self.system_power_control == SystemPowerControl.SHUTDOWN:
            log.info("system will attempt to shutdown now!")
            if platform is tools.Platform.Linux:
                command = "systemctl poweroff"
            else:
                command = "shutdown -h now"
        elif self.system_power_control == SystemPowerControl.SUSPEND:
            log.info("system will attempt to suspend now!")
            if platform is tools.Platform.Linux:
                command = "systemctl suspend"

        if command:
            os.system(command)

    def _on_player_changed(self, event, _):
        if event == "chapter-changed":
            self.stop_after_chapter = False
            self._notify("stop_after_chapter")
        elif event == "play":
            self._start_timer()
        elif event == "pause" or event == "stop":
            self._stop_timer()
        elif event == "fadeout-finished" and self._wait_for_fadeout_end:
            self._wait_for_fadeout_end = False
            self._handle_system_power_event()



from typing import Optional

from cozy.architecture.event_sender import EventSender
from cozy.architecture.observable import Observable
from cozy.ext import inject
from cozy.media.player import Player
from cozy.model.book import Book
from cozy.open_view import OpenView

NS_TO_SEC = 10 ** 9


class PlaybackControlViewModel(Observable, EventSender):
    _player: Player = inject.attr(Player)

    def __init__(self):
        super().__init__()
        super(Observable, self).__init__()

        self._book: Optional[Book] = None

        self._player.add_listener(self._on_player_event)

        if self._player.loaded_book:
            self.book = self._player.loaded_book

    @property
    def book(self) -> Optional[Book]:
        return self._book

    @book.setter
    def book(self, value: Optional[Book]):
        if self._book:
            self._book.remove_bind("playback_speed", self._on_playback_speed_changed)

        self._book = value
        if value:
            self._book.bind_to("playback_speed", self._on_playback_speed_changed)

        self._notify("lock_ui")

    @property
    def playing(self) -> bool:
        if not self._player.loaded_book:
            return False

        return self._player.playing

    @property
    def position(self) -> Optional[float]:
        if not self._book:
            return None

        position = self._book.current_chapter.position - self._book.current_chapter.start_position
        return position / NS_TO_SEC / self._book.playback_speed

    @position.setter
    def position(self, new_value: int):
        if not self._book:
            return

        self._player.position = new_value * self._book.playback_speed

    @property
    def length(self) -> Optional[float]:
        if not self._player.loaded_book or not self._book:
            return None

        return self._player.loaded_book.current_chapter.length / self._book.playback_speed

    @property
    def lock_ui(self) -> bool:
        return not self._book

    @property
    def volume(self) -> float:
        return self._player.volume

    @volume.setter
    def volume(self, new_value: float):
        self._player.volume = new_value

    def play_pause(self):
        self._player.play_pause()

    def rewind(self):
        self._player.rewind()

    def forward(self):
        self._player.forward()

    def open_book_detail(self):
        if self.book:
            self.emit_event(OpenView.BOOK, self.book)

    def _on_player_event(self, event, message):
        if event == "play" or event == "pause":
            if self.book:
                self._notify("playing")
        elif event == "position":
            if self.book:
                self._notify("position")
                self._notify("progress_percent")
                self._notify("remaining_text")
        elif event == "chapter-changed":
            if message:
                self.book = message
                self._notify("book")
                self._notify("position")
                self._notify("length")
                self._notify("volume")
        elif event == "stop":
            self.book = None
            self._notify("book")
            self._notify("position")
            self._notify("length")

    def _on_playback_speed_changed(self):
        self._notify("position")
        self._notify("length")



import locale
import logging
import os
import sys
import threading
from pathlib import Path
from traceback import format_exception

import distro

import gi

from cozy.db.storage import Storage
from cozy.ui.widgets.filter_list_box import FilterListBox
from cozy.ui.widgets.seek_bar import SeekBar

gi.require_version('Handy', '1')

from gi.repository import Gtk, GLib, Handy

from cozy.app_controller import AppController
from cozy.control.db import init_db
from cozy.control.mpris import MPRIS
from cozy.db.settings import Settings
from cozy.report import reporter
from cozy.ui.main_view import CozyUI
from cozy.version import __version__


log = logging.getLogger("application")


def setup_thread_excepthook():
    """
    Workaround for `sys.excepthook` thread bug from:
    http://bugs.python.org/issue1230540

    Call once from the main thread before creating any threads.
    """

    init_original = threading.Thread.__init__

    def init(self, *args, **kwargs):

        init_original(self, *args, **kwargs)
        run_original = self.run

        def run_with_except_hook(*args2, **kwargs2):
            try:
                run_original(*args2, **kwargs2)
            except Exception:
                sys.excepthook(*sys.exc_info())

        self.run = run_with_except_hook

    threading.Thread.__init__ = init


class Application(Gtk.Application):
    ui: CozyUI
    app_controller: AppController

    def __init__(self, localedir: str, pkgdatadir: str):
        self.localedir = localedir
        self.pkgdatadir = pkgdatadir

        Gtk.Application.__init__(self, application_id='com.github.geigi.cozy')
        self.init_custom_widgets()

        GLib.setenv("PULSE_PROP_media.role", "music", True)
        GLib.set_application_name("Cozy")

        self.old_except_hook = sys.excepthook
        sys.excepthook = self.handle_exception
        setup_thread_excepthook()

        import gettext
        locale.bindtextdomain('com.github.geigi.cozy', localedir)
        locale.textdomain('com.github.geigi.cozy')
        gettext.install('com.github.geigi.cozy', localedir)

    def do_startup(self):
        log.info(distro.linux_distribution(full_distribution_name=False))
        log.info("Starting up cozy " + __version__)
        self.ui = CozyUI(self.pkgdatadir, self, __version__)
        init_db()
        Gtk.Application.do_startup(self)
        Handy.init()
        log.info("libhandy version: {}".format(Handy._version))
        self.ui.startup()

    def do_activate(self):
        main_window_builder = self.ui.get_builder()
        self.app_controller = AppController(self, main_window_builder, self.ui)

        self.ui.activate(self.app_controller.library_view)

        if Settings.get().first_start:
            Settings.update(first_start=False).execute()

            path = os.path.join(Path.home(), _("Audiobooks"))
            Storage.create(path=path, default=True)

            os.makedirs(path, exist_ok=True)

        self.add_window(self.ui.window)
        mpris = MPRIS(self)
        mpris._on_current_changed()

    def handle_exception(self, exc_type, exc_value, exc_traceback):
        print("handle exception")
        try:
            reporter.exception("uncaught", exc_value, "\n".join(format_exception(exc_type, exc_value, exc_traceback)))
        except:
            pass

        self.old_except_hook(exc_type, exc_value, exc_traceback)

    def quit(self):
        self.app_controller.quit()
        super(Application, self).quit()


    @staticmethod
    def init_custom_widgets():
        FilterListBox()
        SeekBar()



import gi

from cozy.ext import inject
from cozy.view_model.sleep_timer_view_model import SleepTimerViewModel, SystemPowerControl

gi.require_version('Gtk', '3.0')
from gi.repository import Gtk


@Gtk.Template.from_resource('/com/github/geigi/cozy/timer_popover.ui')
class SleepTimer(Gtk.Popover):
    __gtype_name__ = "SleepTimer"

    _view_model = inject.attr(SleepTimerViewModel)

    timer_scale: Gtk.Scale = Gtk.Template.Child()
    timer_label: Gtk.Label = Gtk.Template.Child()
    timer_grid: Gtk.Grid = Gtk.Template.Child()
    min_label: Gtk.Label = Gtk.Template.Child()
    chapter_switch: Gtk.Switch = Gtk.Template.Child()

    power_control_switch: Gtk.Switch = Gtk.Template.Child()
    power_control_options: Gtk.Box = Gtk.Template.Child()
    system_shutdown_radiob: Gtk.RadioButton = Gtk.Template.Child()
    system_suspend_radiob: Gtk.RadioButton = Gtk.Template.Child()

    def __init__(self, timer_image: Gtk.Image):
        super().__init__()

        self._timer_image: Gtk.Image = timer_image

        self._init_timer_scale()
        self._connect_widgets()

        self._connect_view_model()

        self._on_timer_scale_changed(self.timer_scale)

    def _connect_widgets(self):
        self.timer_scale.connect("value-changed", self._on_timer_scale_changed)
        self.chapter_switch.connect("state-set", self._on_chapter_switch_changed)
        self.power_control_switch.connect("state-set", self._on_power_options_switch_changed)
        self.system_suspend_radiob.connect("toggled", self._on_system_action_radio_button_changed)
        self.system_shutdown_radiob.connect("toggled", self._on_system_action_radio_button_changed)

    def _connect_view_model(self):
        self._view_model.bind_to("stop_after_chapter", self._on_stop_after_chapter_changed)
        self._view_model.bind_to("remaining_seconds", self._on_remaining_seconds_changed)
        self._view_model.bind_to("timer_enabled", self._on_timer_enabled_changed)

    def _init_timer_scale(self):
        for i in range(0, 121, 30):
            self.timer_scale.add_mark(i, Gtk.PositionType.RIGHT, None)

    def _on_timer_scale_changed(self, scale: Gtk.Scale):
        value = scale.get_value()

        if value > 0:
            self.timer_label.set_visible(True)
            self.min_label.set_text(_("min"))
            text = str(int(value))
            self.timer_label.set_text(text)
            self._view_model.remaining_seconds = value * 60
        else:
            self.min_label.set_text(_("Off"))
            self.timer_label.set_visible(False)
            self._view_model.remaining_seconds = 0

    def _on_chapter_switch_changed(self, _, state):
        self.timer_grid.set_sensitive(not state)
        self._view_model.stop_after_chapter = state

    def _on_remaining_seconds_changed(self):
        if self._view_model.remaining_seconds < 1:
            value = 0
        else:
            value = int((self._view_model.remaining_seconds / 60)) + 1

        self.timer_scale.set_value(value)

    def _on_power_options_switch_changed(self, _, state):
        self.power_control_options.set_sensitive(state)

        if not state:
            self._view_model.system_power_control = SystemPowerControl.OFF
        else:
            self._on_system_action_radio_button_changed(None)

    def _on_system_action_radio_button_changed(self, _):
        if self.system_suspend_radiob.get_active():
            self._view_model.system_power_control = SystemPowerControl.SUSPEND
        else:
            self._view_model.system_power_control = SystemPowerControl.SHUTDOWN

    def _on_stop_after_chapter_changed(self):
        self.chapter_switch.set_active(self._view_model.stop_after_chapter)

    def _on_timer_enabled_changed(self):
        if self._view_model.timer_enabled:
            icon = "timer-on-symbolic"
        else:
            icon = "timer-off-symbolic"

        self._timer_image.set_from_icon_name(icon, Gtk.IconSize.BUTTON)



import logging

import gi

from cozy.ext import inject
from cozy.ui.widgets.progress_popover import ProgressPopover
from cozy.view_model.headerbar_view_model import HeaderbarViewModel, HeaderBarState

gi.require_version('Gtk', '3.0')
gi.require_version('Dazzle', '1.0')
from gi.repository import Gtk, Handy
from gi.repository.Handy import HeaderBar
from gi.repository.Dazzle import ProgressMenuButton

log = logging.getLogger("Headerbar")

COVER_SIZE = 45


@Gtk.Template.from_resource('/com/github/geigi/cozy/headerbar.ui')
class Headerbar(HeaderBar):
    __gtype_name__ = "Headerbar"

    search_button: Gtk.MenuButton = Gtk.Template.Child()
    menu_button: Gtk.MenuButton = Gtk.Template.Child()

    progress_menu_button: ProgressMenuButton = Gtk.Template.Child()

    back_button: Gtk.Button = Gtk.Template.Child()
    category_toolbar: Handy.ViewSwitcherTitle = Gtk.Template.Child()

    def __init__(self, main_window_builder: Gtk.Builder):
        super().__init__()

        self._library_mobile_view_switcher: Handy.ViewSwitcherBar = main_window_builder.get_object(
            "library_mobile_view_switcher")
        self._library_mobile_revealer: Gtk.Revealer = main_window_builder.get_object("library_mobile_revealer")
        self._header_container: Gtk.Box = main_window_builder.get_object("header_container")
        self._header_container.pack_start(self, False, True, 0)

        self._sort_stack: Gtk.Stack = main_window_builder.get_object("sort_stack")
        self.category_toolbar.set_stack(self._sort_stack)
        self._library_mobile_view_switcher.set_stack(self._sort_stack)

        self.progress_popover = ProgressPopover()
        self.progress_menu_button.set_popover(self.progress_popover)

        self._headerbar_view_model: HeaderbarViewModel = inject.instance(HeaderbarViewModel)
        self._init_app_menu()
        self._connect_view_model()
        self._connect_widgets()

    def _connect_view_model(self):
        self._headerbar_view_model.bind_to("state", self._on_state_changed)
        self._headerbar_view_model.bind_to("work_progress", self._on_work_progress_changed)
        self._headerbar_view_model.bind_to("work_message", self._on_work_message_changed)
        self._headerbar_view_model.bind_to("can_navigate_back", self._on_can_navigate_back_changed)
        self._headerbar_view_model.bind_to("show_library_filter", self._on_show_library_filter_changed)
        self._headerbar_view_model.bind_to("lock_ui", self._on_lock_ui_changed)

    def _connect_widgets(self):
        self.back_button.connect("clicked", self._back_clicked)
        self.category_toolbar.connect("notify::title-visible", self._on_title_visible_changed)

    def _init_app_menu(self):
        self.menu_builder = Gtk.Builder.new_from_resource("/com/github/geigi/cozy/titlebar_menu.ui")
        menu = self.menu_builder.get_object("titlebar_menu")
        self.menu_button.set_menu_model(menu)

    def _on_state_changed(self):
        if self._headerbar_view_model.state == HeaderBarState.PLAYING:
            progress_visible = False
            self.progress_menu_button.set_progress(0)
        else:
            progress_visible = True

        self.progress_menu_button.set_visible(progress_visible)

    def _on_work_progress_changed(self):
        progress = self._headerbar_view_model.work_progress
        self.progress_menu_button.set_progress(progress)
        self.progress_popover.set_progress(progress)

    def _on_work_message_changed(self):
        self.progress_popover.set_message(self._headerbar_view_model.work_message)

    def _on_can_navigate_back_changed(self):
        self.back_button.set_visible(self._headerbar_view_model.can_navigate_back)

    def _on_show_library_filter_changed(self):
        self.category_toolbar.set_visible(self._headerbar_view_model.show_library_filter)
        self._reveal_mobile_library_filter(self.category_toolbar.get_title_visible())

    def _back_clicked(self, _):
        self._headerbar_view_model.navigate_back()

    def _on_title_visible_changed(self, widget, param):
        visible = widget.get_property(param.name)
        self._reveal_mobile_library_filter(visible)

    def _reveal_mobile_library_filter(self, reveal: bool):
        reveal_child = reveal and self._headerbar_view_model.show_library_filter
        self._library_mobile_revealer.set_reveal_child(reveal_child)

    def _on_lock_ui_changed(self):
        self.search_button.set_sensitive(not self._headerbar_view_model.lock_ui)



import os

import requests
import datetime
import pytz
import distro
import platform

from cozy.application_settings import ApplicationSettings
from cozy.ext import inject
from cozy.report.log_level import LogLevel
from cozy.version import __version__ as CozyVersion
from peewee import __version__ as PeeweeVersion
from mutagen import version_string as MutagenVersion

import gi

gi.require_version('Gtk', '3.0')

from gi.repository import Gtk

URL = 'https://errors.cozy.sh:3100/api/prom/push'
ENABLE = 'true'

LOG_LEVEL_MAP = {
    LogLevel.DEBUG: "DEBUG",
    LogLevel.INFO: "INFO",
    LogLevel.WARNING: "WARN",
    LogLevel.ERROR: "ERROR"
}


def report(component: str, type: LogLevel, message: str, exception: Exception):
    if ENABLE != 'true':
        return

    app_settings = inject.instance(ApplicationSettings)
    report_level = app_settings.report_level

    if report_level == 0:
        return

    curr_datetime = datetime.datetime.now(pytz.timezone('Europe/Berlin'))
    curr_datetime = curr_datetime.isoformat('T')

    if not component or not type or not message:
        raise ValueError("component, type and message are mandatory")

    labels = __append_label("", "component", component)

    if exception:
        labels = __append_label(labels, "exception_type", exception.__class__.__name__)

    labels = __append_label(labels, "app", "cozy")
    labels = __append_label(labels, "level", LOG_LEVEL_MAP[type])

    labels = __append_label(labels, "gtk_version", "{}.{}".format(Gtk.get_major_version(), Gtk.get_minor_version()))
    labels = __append_label(labels, "python_version", platform.python_version())
    labels = __append_label(labels, "peewee_version", PeeweeVersion)
    labels = __append_label(labels, "mutagen_version", MutagenVersion)
    labels = __append_label(labels, "version", CozyVersion)

    if report_level > 1:
        labels = __append_label(labels, "distro", distro.name())
        labels = __append_label(labels, "distro_version", distro.version())
        labels = __append_label(labels, "desktop_environment", os.environ.get('DESKTOP_SESSION'))

    line = "[{}] {}".format(LOG_LEVEL_MAP[type], message)

    headers = {
        'Content-type': 'application/json'
    }
    payload = {
        'streams': [
            {
                'labels': "{{{}}}".format(labels),
                'entries': [
                    {
                        'ts': curr_datetime,
                        'line': line
                    }
                ]
            }
        ]
    }
    try:
        requests.post(URL, json=payload, headers=headers, timeout=10)
    except:
        pass


def __append_label(labels, new_label_name, new_label_content):
    if labels:
        labels += ","
    else:
        labels = ""

    labels += "{}=\"{}\"".format(new_label_name, new_label_content)

    return labels



import math

import cairo
from gi.repository import Gtk, GdkPixbuf, Gdk

RADIUS = 5


class AlbumArt(Gtk.DrawingArea):
    def __init__(self, **properties):
        super().__init__(**properties)

        self.art = None

        self.set_visible(True)
        self.connect("draw", self._on_draw)

    def set_art(self, art: GdkPixbuf):
        self.art = art
        scale = self.get_scale_factor()
        self.set_size_request(art.get_width() / scale, art.get_height() / scale)
        self.queue_draw()

    def _draw_rounded_path(self, context: cairo.Context):
        degrees = math.pi / 180.0

        context.new_sub_path()
        context.arc(self.width - RADIUS, RADIUS, RADIUS, -90 * degrees, 0 * degrees)
        context.arc(self.width - RADIUS, self.height - RADIUS, RADIUS, 0 * degrees, 90 * degrees)
        context.arc(RADIUS, self.height - RADIUS, RADIUS, 90 * degrees, 180 * degrees)
        context.arc(RADIUS, RADIUS, RADIUS, 180 * degrees, 270 * degrees)
        context.close_path()

    def _on_draw(self, area: Gtk.DrawingArea, context: cairo.Context):
        self.height = area.get_allocated_height()
        self.width = area.get_allocated_width()

        self._draw_rounded_path(context)

        if self.art:
            surface = Gdk.cairo_surface_create_from_pixbuf(self.art, self.get_scale_factor(), None)
            context.set_source_surface(surface, 0, 0)
            context.clip()

        context.paint()
        return False



import logging
import uuid
import os


from cozy.architecture.event_sender import EventSender
from cozy.control.application_directories import get_cache_dir
import cozy.tools as tools
import cozy.ui

from gi.repository import Gio

from cozy.db.file import File
from cozy.db.offline_cache import OfflineCache as OfflineCacheModel
from cozy.db.track_to_file import TrackToFile
from cozy.ext import inject
from cozy.model.book import Book
from cozy.model.chapter import Chapter
from cozy.report import reporter

log = logging.getLogger("offline_cache")


class OfflineCache(EventSender):
    """
    This class is responsible for all actions on the offline cache.
    This includes operations like copying to the cache and adding or removing files from
    the cache.
    """
    queue = []
    total_batch_count = 0
    current_batch_count = 0
    current = None
    thread = None
    filecopy_cancel = None
    last_ui_update = 0
    current_book_processing = None

    def __init__(self):
        super().__init__()

        from cozy.media.importer import Importer
        self._importer = inject.instance(Importer)

        from cozy.model.library import Library
        self._library = inject.instance(Library)

        self._importer.add_listener(self._on_importer_event)

        self.cache_dir = os.path.join(get_cache_dir(), "offline")
        if not os.path.exists(self.cache_dir):
            os.makedirs(self.cache_dir)

        self._start_processing()

        inject.instance(cozy.ui.settings.Settings).add_listener(self.__on_settings_changed)

    def add(self, book: Book):
        """
        Add all tracks of a book to the offline cache and start copying.
        """
        files_to_cache = []
        file_ids = {chapter.file_id for chapter in book.chapters}
        for file_id in file_ids:
            cached_file_name = str(uuid.uuid4())
            files_to_cache.append((file_id, cached_file_name))
        chunks = [files_to_cache[x:x + 500] for x in range(0, len(files_to_cache), 500)]
        for chunk in chunks:
            query = OfflineCacheModel.insert_many(chunk, fields=[OfflineCacheModel.original_file,
                                                                 OfflineCacheModel.cached_file])
            self.total_batch_count += len(chunk)
            query.execute()

        self._start_processing()

    def remove(self, book: Book):
        """
        Remove all tracks of the given book from the cache.
        """
        self._stop_processing()
        ids = {t.file_id for t in book.chapters}
        offline_elements = OfflineCacheModel.select().join(File).where(OfflineCacheModel.original_file.id << ids)

        for element in offline_elements:
            file_path = os.path.join(self.cache_dir, element.cached_file)
            if file_path == self.cache_dir:
                continue

            file = Gio.File.new_for_path(file_path)
            if file.query_exists():
                file.delete()

            for item in self.queue:
                if self.current and item.id == self.current.id:
                    self.filecopy_cancel.cancel()

        entries_to_delete = OfflineCacheModel.select().join(File).where(OfflineCacheModel.original_file.id << ids)
        ids_to_delete = [t.id for t in entries_to_delete]
        OfflineCacheModel.delete().where(OfflineCacheModel.id << ids_to_delete).execute()
        book.downloaded = False
        self.emit_event("book-offline-removed", book)
        self.queue = []

        self._start_processing()

    def remove_all_for_storage(self, storage_path):
        """
        """
        for element in OfflineCacheModel.select().join(File).where(
                storage_path in OfflineCacheModel.original_file.path):
            file_path = os.path.join(self.cache_dir, element.cached_file)
            if file_path == self.cache_dir:
                continue

            file = Gio.File.new_for_path(file_path)
            if file.query_exists():
                file.delete()

            if element.track.book.offline == True:
                element.track.book.update(offline=False, downloaded=False).execute()

        OfflineCacheModel.delete().where(storage_path in OfflineCacheModel.original_file.path).execute()

    def get_cached_path(self, chapter: Chapter):
        query = OfflineCacheModel.select().where(OfflineCacheModel.original_file == chapter.file_id,
                                                 OfflineCacheModel.copied == True)
        if query.count() > 0:
            return os.path.join(self.cache_dir, query.get().cached_file)
        else:
            return None

    def update_cache(self, paths):
        """
        Update the cached version of the given files.
        """
        if OfflineCacheModel.select().count() > 0:
            OfflineCacheModel.update(copied=False).where(
                OfflineCacheModel.original_file.path in paths).execute()
            self._fill_queue_from_db()

    def delete_cache(self):
        """
        Deletes the entire offline cache files.
        Doesn't delete anything from the cozy.db.
        """
        cache_dir = os.path.join(get_cache_dir(), "offline")

        import shutil
        shutil.rmtree(cache_dir)

    def _stop_processing(self):
        """
        """
        if not self._is_processing() or not self.thread:
            return

        self.filecopy_cancel.cancel()
        self.thread.stop()

    def _start_processing(self):
        """
        """
        if self._is_processing():
            return

        self.thread = tools.StoppableThread(target=self._process_queue)
        self.thread.start()

    def _process_queue(self):
        log.info("Started processing queue")
        self.filecopy_cancel = Gio.Cancellable()

        self._fill_queue_from_db()
        self.total_batch_count = len(self.queue)
        self.current_batch_count = 0
        if len(self.queue) > 0:
            self.current_book_processing = self._get_book_to_file(self.queue[0].original_file).id
            self.emit_event_main_thread("start")

        while len(self.queue) > 0:
            log.info("Processing item")
            self.current_batch_count += 1
            item = self.queue[0]
            if self.thread.stopped():
                break

            query = OfflineCacheModel.select().where(OfflineCacheModel.id == item.id)
            if not query.exists():
                continue

            new_item = OfflineCacheModel.get(OfflineCacheModel.id == item.id)

            book = self._get_book_to_file(new_item.original_file)
            if self.current_book_processing != book.id:
                self._update_book_download_status(self.current_book_processing)
                self.current_book_processing = book.id

            if not new_item.copied and os.path.exists(new_item.original_file.path):
                log.info("Copying item")
                self.emit_event_main_thread("message",
                                            _("Copying") + " " + tools.shorten_string(book.name, 30))
                self.current = new_item

                destination = Gio.File.new_for_path(os.path.join(self.cache_dir, new_item.cached_file))
                source = Gio.File.new_for_path(new_item.original_file.path)
                flags = Gio.FileCopyFlags.OVERWRITE
                try:
                    copied = source.copy(destination, flags, self.filecopy_cancel, self.__update_copy_status, None)
                except Exception as e:
                    if e.code == Gio.IOErrorEnum.CANCELLED:
                        log.info("Download of book was cancelled.")
                        self.thread.stop()
                        break
                    reporter.exception("offline_cache", e)
                    log.error("Could not copy file to offline cache: " + new_item.original_file.path)
                    log.error(e)
                    self.queue.remove(item)
                    continue

                if copied:
                    OfflineCacheModel.update(copied=True).where(
                        OfflineCacheModel.id == new_item.id).execute()

            self.queue.remove(item)

        if self.current_book_processing:
            self._update_book_download_status(self.current_book_processing)

        self.current = None
        self.emit_event_main_thread("finished")

    def _get_book_to_file(self, file: File):
        track_to_file = TrackToFile.select().join(File).where(TrackToFile.file == file.id).get()
        return track_to_file.track.book

    def _update_book_download_status(self, book_id):
        book = next(book for book in self._library.books if book.id == book_id)
        downloaded = self._is_book_downloaded(book)

        book.downloaded = downloaded

        if downloaded:
            self.emit_event("book-offline", book)
        else:
            self.emit_event("book-offline-removed", book)

    def _is_book_downloaded(self, book: Book):
        file_ids = [chapter.file_id for chapter in book.chapters]
        offline_files = OfflineCacheModel.select().where(OfflineCacheModel.original_file << file_ids)
        offline_file_ids = [file.original_file.id for file in offline_files]

        for chapter in book.chapters:
            if chapter.file_id not in offline_file_ids:
                return False

        return True

    def _is_processing(self):
        """
        """
        if self.thread:
            return self.thread.is_alive()
        else:
            return False

    def _fill_queue_from_db(self):
        for item in OfflineCacheModel.select().where(OfflineCacheModel.copied == False):
            if not any(item.id == queued.id for queued in self.queue):
                self.queue.append(item)
                self.total_batch_count += 1

    def _on_importer_event(self, event: str, message):
        if event == "new-or-updated-files":
            self.update_cache(message)
            self._start_processing()

    def __update_copy_status(self, current_num_bytes, total_num_bytes, _):
        total_batch_count = max(self.total_batch_count, 1)

        finished_files_progress = (self.current_batch_count - 1) / total_batch_count
        current_file_progress = current_num_bytes / max(total_num_bytes, 1)

        progress = finished_files_progress + (current_file_progress / total_batch_count)
        self.emit_event_main_thread("progress", min(progress, 1))

    def __on_settings_changed(self, event, message):
        """
        This method reacts to storage settings changes.
        """
        if event == "storage-removed" or event == "external-storage-removed":
            self.remove_all_for_storage(message)



import os
import uuid
import logging

from gi.repository import GdkPixbuf

from cozy.application_settings import ApplicationSettings
from cozy.control.application_directories import get_cache_dir
from cozy.db.artwork_cache import ArtworkCache as ArtworkCacheModel
from cozy.ext import inject
from cozy.media.importer import Importer, ScanStatus
from cozy.report import reporter

log = logging.getLogger("artwork_cache")


class ArtworkCache:
    def __init__(self):
        _importer = inject.instance(Importer)
        _importer.add_listener(self._on_importer_event)

    def get_cover_pixbuf(self, book, scale, size=0):
        pixbuf = None
        size *= scale

        if size > 0:
            # first try the cache
            pixbuf = self._load_pixbuf_from_cache(book, size)

        if pixbuf:
            return pixbuf
        else:
            # then try the db or file
            pixbuf = self._load_cover_pixbuf(book)

        if pixbuf:
            # return original size if it is not greater than 0
            if not size > 0:
                return pixbuf

            # create cached version
            pixbuf = self._create_artwork_cache(book, pixbuf, size)
        else:
            pixbuf = None

        return pixbuf

    def delete_artwork_cache(self):
        """
        Deletes the artwork cache completely.
        """
        cache_dir = os.path.join(get_cache_dir(), "artwork")

        import shutil
        if os.path.exists(cache_dir):
            shutil.rmtree(cache_dir)

        q = ArtworkCacheModel.delete()
        q.execute()

    def _on_importer_event(self, event, data):
        if event == "scan":
            if data == ScanStatus.STARTED:
                self.delete_artwork_cache()

    def _create_artwork_cache(self, book, pixbuf, size):
        """
        Creates a resized cache version of the given pixbuf and saves it
        in the cozy cache folder under a unique identifier.
        :param book: Book which the artwork is from
        :param pixbuf: Pixbuf to be cached
        :param size: Size for the cached version
        :return: Resized pixbuf
        """
        query = ArtworkCacheModel.select().where(ArtworkCacheModel.book == book.id)
        gen_uuid = ""

        if query.exists():
            gen_uuid = str(query.first().uuid)
        else:
            gen_uuid = str(uuid.uuid4())
            ArtworkCacheModel.create(book=book.id, uuid=gen_uuid)

        cache_dir = os.path.join(os.path.join(get_cache_dir(), "artwork"), gen_uuid)
        if not os.path.exists(cache_dir):
            os.makedirs(cache_dir)

        resized_pixbuf = self._resize_pixbuf(pixbuf, size)
        file_path = os.path.join(cache_dir, str(size) + ".jpg")
        if not os.path.exists(file_path):
            try:
                resized_pixbuf.savev(file_path, "jpeg", ["quality", None], ["95"])
            except Exception as e:
                reporter.warning("artwork_cache", "Failed to save resized cache albumart")
                log.warning("Failed to save resized cache albumart for following uuid: " + gen_uuid)
                log.warning(e)

        return resized_pixbuf

    def get_album_art_path(self, book, size):
        query = ArtworkCacheModel.select().where(ArtworkCacheModel.book == book.id)
        if query.exists():
            try:
                uuid = query.first().uuid
            except Exception as e:
                reporter.error("artwork_cache", "load_pixbuf_from_cache: query exists but query.first().uuid crashed.")
                return None
        else:
            return None

        cache_dir = os.path.join(get_cache_dir(), "artwork")
        cache_dir = os.path.join(cache_dir, uuid)

        try:
            if os.path.exists(cache_dir):
                file_path = os.path.join(cache_dir, str(size) + ".jpg")
                if os.path.exists(file_path):
                    return file_path
                else:
                    return None
        except Exception as e:
            log.warning(e)
            return None

        return None

    def _load_pixbuf_from_cache(self, book, size):
        path = self.get_album_art_path(book, size)

        try:
            if path:
                pixbuf = GdkPixbuf.Pixbuf.new_from_file(path)
            else:
                pixbuf = None
        except Exception as e:
            log.warning("Failed to load pixbuf from path: {}. Deleting file.".format(path))
            log.debug(e)
            os.remove(path)
            return None

        return pixbuf

    @inject.autoparams()
    def _load_cover_pixbuf(self, book, app_settings: ApplicationSettings):
        """
        Load the cover from a given book and create a pixbuf object with a given from it.
        :param book: The book object
        :param size: The size of the bigger side in pixels
        :return: pixbuf object containing the cover
        """
        pixbuf = None

        if app_settings.prefer_external_cover:
            pixbuf = self._load_pixbuf_from_file(book)

            if pixbuf is None:
                pixbuf = self._load_pixbuf_from_db(book)
        else:
            pixbuf = self._load_pixbuf_from_db(book)

            if pixbuf is None:
                pixbuf = self._load_pixbuf_from_file(book)

        return pixbuf

    def _load_pixbuf_from_db(self, book):
        pixbuf = None

        if book and book.cover:
            try:
                loader = GdkPixbuf.PixbufLoader.new()
                loader.write(book.cover)
                loader.close()
                pixbuf = loader.get_pixbuf()
            except Exception as e:
                reporter.warning("artwork_cache", "Could not get book cover from db.")
                log.warning("Could not get cover for book " + book.name)
                log.warning(e)

        return pixbuf

    def _resize_pixbuf(self, pixbuf, size):
        """
        Resizes an pixbuf and keeps the aspect ratio.
        :return: Resized pixbuf.
        """
        resized_pixbuf = pixbuf

        if size > 0:
            if pixbuf.get_height() > pixbuf.get_width():
                width = int(pixbuf.get_width() / (pixbuf.get_height() / size))
                resized_pixbuf = pixbuf.scale_simple(
                    width, size, GdkPixbuf.InterpType.BILINEAR)
            else:
                height = int(pixbuf.get_height() / (pixbuf.get_width() / size))
                resized_pixbuf = pixbuf.scale_simple(
                    size, height, GdkPixbuf.InterpType.BILINEAR)

        return resized_pixbuf

    def _load_pixbuf_from_file(self, book):
        """
        Try to load the artwork from a book from image files.
        :param book: The book to load the artwork from.
        :return: Artwork as pixbuf object.
        """
        pixbuf = None
        cover_files = []

        try:
            directory = os.path.dirname(os.path.normpath(book.chapters[0].file))

            cover_files = [f for f in os.listdir(directory)
                           if f.lower().endswith('.png') or f.lower().endswith(".jpg") or f.lower().endswith(".gif")]
        except Exception as e:
            log.warning("Could not open audiobook directory and look for cover files.")
            log.warning(e)
        for elem in (x for x in cover_files if os.path.splitext(x.lower())[0] == "cover"):
            # find cover.[jpg,png,gif]
            try:
                pixbuf = GdkPixbuf.Pixbuf.new_from_file(os.path.join(directory, elem))
            except Exception as e:
                log.debug(e)
            if pixbuf:
                break
        if pixbuf is None:
            # find other cover file (sort alphabet)
            cover_files.sort(key=str.lower)
            for elem in cover_files:
                try:
                    pixbuf = GdkPixbuf.Pixbuf.new_from_file(os.path.join(directory, elem))
                except Exception as e:
                    log.debug(e)
                if pixbuf:
                    break
        return pixbuf



from peewee import ForeignKeyField, IntegerField

from cozy.db.file import File
from cozy.db.model_base import ModelBase
from cozy.db.track import Track


class TrackToFile(ModelBase):
    track = ForeignKeyField(Track, unique=True, backref='track_to_file')
    file = ForeignKeyField(File, unique=False)
    start_at = IntegerField()



import logging
from typing import List

import cozy.ext.inject as inject
from gi.repository import Gio

from cozy.architecture.event_sender import EventSender
from cozy.model.book import Book
from cozy.model.settings import Settings
from cozy.model.storage import Storage
from cozy.report import reporter

log = logging.getLogger("fs_monitor")


class StorageNotFound(Exception):
    pass


class ExternalStorage:
    storage: Storage
    online: bool

    def __init__(self, storage: Storage, online: bool):
        self.storage = storage
        self.online = online


class FilesystemMonitor(EventSender):
    external_storage: List[ExternalStorage] = []
    _settings: Settings = inject.attr(Settings)

    def __init__(self):
        super().__init__()
        self.volume_monitor: Gio.VolumeMonitor = Gio.VolumeMonitor.get()
        self.volume_monitor.connect("mount-added", self.__on_mount_added)
        self.volume_monitor.connect("mount-removed", self.__on_mount_removed)

        self.init_offline_mode()

        from cozy.ui.settings import Settings as UISettings
        self._ui_settings = inject.instance(UISettings)
        self._ui_settings.add_listener(self.__on_settings_changed)

    def init_offline_mode(self):
        external_storage = []
        mounts = self.volume_monitor.get_mounts()
        # go through all audiobook locations and test if they can be found in the mounts list

        for storage in self._settings.external_storage_locations:
            online = False
            if any(mount.get_root().get_path() in storage.path for mount in mounts):
                online = True
            self.external_storage.append(ExternalStorage(storage=storage, online=online))

    def close(self):
        """
        Free all references.
        """
        # self.volume_monitor.unref()
        pass

    def get_book_online(self, book: Book):
        result = next(
            (storage.online for storage in self.external_storage if storage.storage.path in book.chapters[0].file),
            True)
        return result

    def is_track_online(self, track):
        """
        """
        result = next((storage.online for storage in self.external_storage if storage.storage.path in track.file), True)
        return (result)

    def get_offline_storages(self):
        return [i.storage.path for i in self.external_storage if not i.online]

    def is_storage_online(self, storage: Storage) -> bool:
        storage = next((storage for storage in self.external_storage if storage.storage == storage), None)

        if not storage:
            raise StorageNotFound

        return storage.online

    def is_external(self, directory: str) -> bool:
        mounts: List[Gio.Mount] = self.volume_monitor.get_mounts()

        for mount in mounts:
            root = mount.get_root()
            if not root:
                log.error("Failed to test for external drive. Mountpoint has no root object.")
                reporter.error("fs_monitor", "Failed to test for external drive. Mountpoint has no root object.")
                return False

            path = root.get_path()
            if not path:
                log.error("Failed to test for external drive. Root object has no path.")
                reporter.error("fs_monitor", "Failed to test for external drive. Root object has no path.")
                return False

            if path in directory and mount.can_unmount():
                return True

        return False

    def __on_mount_added(self, monitor, mount):
        """
        A volume was mounted.
        Disable offline mode for this volume.
        """
        mount_path = mount.get_root().get_path()

        if not mount_path:
            log.warning("Mount added but no mount_path is present. Skipping...")
            return

        log.debug("Volume mounted: " + mount_path)

        storage = next((s for s in self.external_storage if mount_path in s.storage.path), None)
        if storage:
            log.info("Storage online: " + mount_path)
            storage.online = True
            self.emit_event("storage-online", storage.storage.path)

    def __on_mount_removed(self, monitor, mount):
        """
        A volume was unmounted.
        Enable offline mode for this volume.
        """
        mount_path = mount.get_root().get_path()

        if not mount_path:
            log.warning("Mount removed but no mount_path is present. Skipping...")
            return

        log.debug("Volume unmounted: " + mount_path)

        storage = next((s for s in self.external_storage if mount_path in s.storage.path), None)
        if storage:
            log.info("Storage offline: " + mount_path)
            storage.online = False
            self.emit_event("storage-offline", storage.storage.path)

            # switch to offline version if currently playing

    def __on_settings_changed(self, event, message):
        """
        This method reacts to storage settings changes.
        """
        if event == "external-storage-added" or event == "storage-changed" or (
                event == "storage-added" and message != ""):
            self.init_offline_mode()
        elif event == "storage-removed" or event == "external-storage-removed":
            self.external_storage = [item for item in self.external_storage if item.storage.path not in message]



import logging
import gi

from cozy.db.storage import Storage
from cozy.db.storage_blacklist import StorageBlackList
from cozy.ext import inject
from cozy.model.library import Library
from cozy.ui.widgets.ScrollWrapper import ScrollWrapper
from cozy.ui.widgets.storage_list_box_row import StorageListBoxRow
from cozy.view_model.settings_view_model import SettingsViewModel

gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, Gio

from cozy.architecture.event_sender import EventSender

import cozy.ui

log = logging.getLogger("settings")


class Settings(EventSender):
    """
    This class contains all logic for cozys preferences.
    """
    _glib_settings: Gio.Settings = inject.attr(Gio.Settings)

    view_model = None
    ui = None
    default_dark_mode = None

    def __init__(self):
        super().__init__()
        from cozy.control.artwork_cache import ArtworkCache
        self._artwork_cache: ArtworkCache = inject.instance(ArtworkCache)
        self._library: Library = inject.instance(Library)

        self.view_model = SettingsViewModel()

        self.ui = cozy.ui.main_view.CozyUI()
        self.builder = Gtk.Builder.new_from_resource(
            "/com/github/geigi/cozy/settings.ui")

        # get settings window
        self.window = self.builder.get_object("settings_window")
        self.window.set_modal(self.ui.window)
        self.window.connect("delete-event", self.ui.hide_window)

        self.add_storage_button = self.builder.get_object("add_location_button")
        self.add_storage_button.connect("clicked", self.__on_add_storage_clicked)
        self.remove_storage_button = self.builder.get_object("remove_location_button")
        self.remove_storage_button.connect("clicked", self.__on_remove_storage_clicked)
        self.external_button = self.builder.get_object("external_button")
        self.external_button_handle_id = self.external_button.connect("clicked", self.__on_external_clicked)
        self.default_storage_button = self.builder.get_object("default_location_button")
        self.default_storage_button.connect("clicked", self.__on_default_storage_clicked)
        self.storage_list_box = self.builder.get_object("storage_list_box")
        self.storage_list_box.connect("row-activated", self.__on_storage_box_changed)

        self.remove_blacklist_button = self.builder.get_object("remove_blacklist_button")
        self.remove_blacklist_button.connect("clicked", self.__on_remove_blacklist_clicked)
        # self.remove_blacklist_button.set_sensitive(False)
        self.blacklist_tree_view = self.builder.get_object("blacklist_tree_view")
        self.blacklist_model = self.builder.get_object("blacklist_store")
        self.blacklist_tree_view.get_selection().connect("changed", self.__on_blacklist_selection_changed)

        self.sleep_fadeout_switch = self.builder.get_object("sleep_fadeout_switch")
        self.sleep_fadeout_switch.connect("notify::active", self.__on_fadeout_switch_changed)

        self.external_cover_switch = self.builder.get_object("external_cover_switch")
        self.external_cover_switch.connect("state-set", self.__on_external_cover_switch_changed)

        self.fadeout_duration_label = self.builder.get_object("fadeout_duration_label")
        self.fadeout_duration_row = self.builder.get_object("fadeout_duration_row")

        self.fadeout_duration_adjustment = self.builder.get_object("fadeout_duration_adjustment")
        self.fadeout_duration_adjustment.connect("value-changed", self.__on_fadeout_adjustment_changed)
        self.__on_fadeout_adjustment_changed(self.fadeout_duration_adjustment)

        self.rewind_duration_label = self.builder.get_object("rewind_duration_label")
        self.rewind_duration_scale = self.builder.get_object("rewind_duration_scale")
        self.rewind_duration_adjustment = self.builder.get_object("rewind_duration_adjustment")
        self.rewind_duration_adjustment.connect("value-changed", self._on_rewind_adjustment_changed)
        self._on_rewind_adjustment_changed(self.rewind_duration_adjustment)
        for i in [10, 15, 20, 30, 45, 60, 75, 90, 105]:
            self.rewind_duration_scale.add_mark(i, Gtk.PositionType.RIGHT, None)

        self.forward_duration_label = self.builder.get_object("forward_duration_label")
        self.forward_duration_scale = self.builder.get_object("forward_duration_scale")
        self.forward_duration_adjustment = self.builder.get_object("forward_duration_adjustment")
        self.forward_duration_adjustment.connect("value-changed", self._on_forward_adjustment_changed)
        self._on_forward_adjustment_changed(self.forward_duration_adjustment)
        for i in [10, 15, 20, 30, 45, 60, 75, 90, 105]:
            self.forward_duration_scale.add_mark(i, Gtk.PositionType.RIGHT, None)

        self.force_refresh_button = self.builder.get_object("force_refresh_button")
        self.force_refresh_button.connect("clicked", self.__on_force_refresh_clicked)

        self.settings_stack: Gtk.Stack = self.builder.get_object("settings_stack")
        self.settings_stack.connect("notify::visible-child", self._on_settings_stack_changed)

        from cozy.ui.widgets.error_reporting import ErrorReporting
        self.settings_stack.add_titled(ScrollWrapper(ErrorReporting()), "feedback", _("Feedback"))

        self._init_storage()
        self._init_blacklist()
        self.__init_bindings()
        self.__on_storage_box_changed(None, None)

        self.set_darkmode()

    def _init_storage(self):
        """
        Display settings from the database in the ui.
        """
        found_default = False
        self.storage_list_box.remove_all_children()
        for location in Storage.select():
            row = StorageListBoxRow(self, location.id, location.path, location.external, location.default)
            self.storage_list_box.add(row)
            if location.default:
                if found_default:
                    row.set_default(False)
                else:
                    found_default = True
                    self.storage_list_box.select_row(row)

        self.__on_storage_box_changed(None, None)

    def _init_blacklist(self):
        """
        Init the Storage location list.
        """
        for file in StorageBlackList.select():
            self.blacklist_model.append([file.path, file.id])
        self.__on_blacklist_selection_changed(None)

    def __init_bindings(self):
        """
        Bind Gio.Settings to widgets in settings dialog.
        """
        sl_switch = self.builder.get_object("symlinks_switch")
        self._glib_settings.bind("symlinks", sl_switch, "active",
                                 Gio.SettingsBindFlags.DEFAULT)

        auto_scan_switch = self.builder.get_object("auto_scan_switch")
        self._glib_settings.bind("autoscan", auto_scan_switch,
                                 "active", Gio.SettingsBindFlags.DEFAULT)

        timer_suspend_switch = self.builder.get_object(
            "timer_suspend_switch")
        self._glib_settings.bind("suspend", timer_suspend_switch,
                                 "active", Gio.SettingsBindFlags.DEFAULT)

        replay_switch = self.builder.get_object("replay_switch")
        self._glib_settings.bind("replay", replay_switch, "active",
                                 Gio.SettingsBindFlags.DEFAULT)

        titlebar_remaining_time_switch = self.builder.get_object("titlebar_remaining_time_switch")
        self._glib_settings.bind("titlebar-remaining-time", titlebar_remaining_time_switch, "active",
                                 Gio.SettingsBindFlags.DEFAULT)

        dark_mode_switch = self.builder.get_object("dark_mode_switch")
        self._glib_settings.bind("dark-mode", dark_mode_switch, "active",
                                 Gio.SettingsBindFlags.DEFAULT)

        swap_author_reader_switch = self.builder.get_object("swap_author_reader_switch")
        self._glib_settings.bind("swap-author-reader", swap_author_reader_switch, "active",
                                 Gio.SettingsBindFlags.DEFAULT)

        self._glib_settings.bind("prefer-external-cover", self.external_cover_switch, "active",
                                 Gio.SettingsBindFlags.DEFAULT)

        self._glib_settings.bind("sleep-timer-fadeout", self.sleep_fadeout_switch, "active",
                                 Gio.SettingsBindFlags.DEFAULT)

        self._glib_settings.bind("sleep-timer-fadeout-duration", self.fadeout_duration_adjustment,
                                 "value", Gio.SettingsBindFlags.DEFAULT)

        self._glib_settings.bind("rewind-duration", self.rewind_duration_adjustment, "value", Gio.SettingsBindFlags.DEFAULT)
        self._glib_settings.bind("forward-duration", self.forward_duration_adjustment, "value", Gio.SettingsBindFlags.DEFAULT)

        self._glib_settings.connect("changed", self.__on_settings_changed)

    def show(self):
        """
        Show the settings window.
        """
        self.window.show()

    def get_storage_elements_blocked(self):
        """
        Are the location ui elements currently blocked?
        """
        return not self.storage_list_box.get_sensitive()

    def block_ui_elements(self, block):
        """
        Block/unblock UI storage elements.
        """
        sensitive = not block
        self.storage_list_box.set_sensitive(sensitive)
        self.add_storage_button.set_sensitive(sensitive)
        self.external_button.set_sensitive(sensitive)

        row = self.storage_list_box.get_selected_row()
        if row and row.get_default() != True:
            self.remove_storage_button.set_sensitive(sensitive)
            self.default_storage_button.set_sensitive(sensitive)

    def __on_add_storage_clicked(self, widget):
        """
        Add a new storage selector to the ui.
        """
        db_obj = Storage.create(path="")
        self.storage_list_box.add(StorageListBoxRow(self, db_obj.id, "", False, False))

    def __on_remove_storage_clicked(self, widget):
        """
        Remove a storage selector from the ui and database.
        """
        row = self.storage_list_box.get_selected_row()
        Storage.select().where(Storage.path == row.path).get().delete_instance()
        self.storage_list_box.remove(row)
        self.emit_event("storage-removed", row.path)
        self.__on_storage_box_changed(None, None)

    def __on_default_storage_clicked(self, widget):
        """
        Select a location as default storage.
        """
        for row in self.storage_list_box.get_children():
            row.set_default(False)

        self.storage_list_box.get_selected_row().set_default(True)

        self.__on_storage_box_changed(None, None)

    def __on_storage_box_changed(self, widget, row):
        """
        Disable/enable toolbar buttons
        """
        row = self.storage_list_box.get_selected_row()
        if row is None:
            sensitive = False
            default_sensitive = False
        else:
            sensitive = True
            if row.get_default():
                default_sensitive = False
            else:
                default_sensitive = True
            self.external_button.handler_block(self.external_button_handle_id)
            self.external_button.set_active(row.external)
            self.external_button.handler_unblock(self.external_button_handle_id)

        self.remove_storage_button.set_sensitive(default_sensitive)
        self.external_button.set_sensitive(sensitive)
        self.default_storage_button.set_sensitive(default_sensitive)

        for child in self.storage_list_box.get_children():
            if row and child.db_id == row.db_id:
                child.set_selected(True)
            else:
                child.set_selected(False)

    def __on_settings_changed(self, settings, key):
        """
        Updates cozy's ui to changed Gio settings.
        """
        if key == "dark-mode":
            self.set_darkmode()

    def __on_remove_blacklist_clicked(self, widget):
        """
        Remove the selected storage from the db and ui.
        TODO: Does this trigger a rescan?
        """
        model, pathlist = self.blacklist_tree_view.get_selection().get_selected_rows()
        if pathlist:
            ids = []
            for path in reversed(pathlist):
                treeiter = model.get_iter(path)
                ids.append(self.blacklist_model.get_value(treeiter, 1))
                self.blacklist_model.remove(treeiter)

            StorageBlackList.delete().where(StorageBlackList.id in ids).execute()

        self.__on_blacklist_selection_changed(self.blacklist_tree_view.get_selection())

    def __on_blacklist_selection_changed(self, tree_selection):
        """
        The user selected a different storage location.
        Here we enable or disable the remove button depending on 
        weather this is the default storage or not.
        """
        if tree_selection is None or len(tree_selection.get_selected_rows()[1]) < 1:
            self.remove_blacklist_button.set_sensitive(False)
        else:
            self.remove_blacklist_button.set_sensitive(True)

    def __on_external_clicked(self, widget):
        """
        The external/internal button was clicked.
        The new setting will be written to the cozy.
        """
        external = self.external_button.get_active()

        row = self.storage_list_box.get_selected_row()
        row.set_external(external)

        if external:
            self.emit_event("external-storage-added", row.path)
        else:
            self.emit_event("external-storage-removed", row.path)

    def __on_fadeout_adjustment_changed(self, adjustment):
        """
        This refreshes the label belonging to the fadeout duration adjustment.
        """
        self.fadeout_duration_label.set_text(str(int(adjustment.get_value())) + " s")

    def _on_rewind_adjustment_changed(self, adjustment):
        self.rewind_duration_label.set_text(str(int(adjustment.get_value())) + " s")

    def _on_forward_adjustment_changed(self, adjustment):
        self.forward_duration_label.set_text(str(int(adjustment.get_value())) + " s")

    def __on_fadeout_switch_changed(self, switch, state):
        """
        Enable/Disable sensitivity for the fadeout duration settings row.
        """
        self.fadeout_duration_row.set_sensitive(switch.get_active())

    def __on_external_cover_switch_changed(self, switch, state):
        """
        Set the glib setting prefer-external-cover.
        This is needed because the binding gets called after this function.
        Then refresh the artwork cache.
        """
        # We have to test if everything is initialized before triggering the refresh
        # otherwise this might be just the initial call when starting up
        if self.ui.is_initialized:
            self._glib_settings.set_boolean("prefer-external-cover", state)
            self._artwork_cache.delete_artwork_cache()
            self.emit_event_main_thread("prefer-external-cover-changed")

    def __on_force_refresh_clicked(self, widget):
        """
        Start a force refresh of the database.
        """
        self._library.reset_modified_date_for_all()
        self.ui.scan(None, None)

    def _on_settings_stack_changed(self, widget, property):
        page = widget.props.visible_child_name

        if page == "files":
            self.blacklist_model.clear()
            self._init_blacklist()

    def set_darkmode(self):
        """
        Enable or disable the dark gtk theme.
        """
        settings = Gtk.Settings.get_default()

        if self.default_dark_mode is None:
            self.default_dark_mode = settings.get_property("gtk-application-prefer-dark-theme")

        user_enabled = self._glib_settings.get_boolean("dark-mode")
        if user_enabled:
            settings.set_property("gtk-application-prefer-dark-theme", True)
        else:
            settings.set_property("gtk-application-prefer-dark-theme", self.default_dark_mode)


class BlacklistColumn(Gtk.TreeViewColumn):
    """
    A column for a storage location.
    """

    def __init__(self, path):
        super(Gtk.TreeViewColumn, self).__init__()
        self.path = path



import gi

from cozy.ext import inject

gi.require_version('Gtk', '3.0')
from gi.repository import Gtk


@Gtk.Template(resource_path='/com/github/geigi/cozy/delete_book_dialog.ui')
class DeleteBookView(Gtk.Dialog):
    __gtype_name__ = 'DeleteBookDialog'

    main_window = inject.attr("MainWindow")

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.set_modal(self.main_window.window)

    def get_delete_book(self):
        response = self.run()

        if response == Gtk.ResponseType.APPLY:
            return True
        else:
            return False



import logging
from typing import List, Set, Optional

from peewee import SqliteDatabase

from cozy.architecture.event_sender import EventSender
from cozy.architecture.profiler import timing
from cozy.db.book import Book as BookModel
from cozy.db.file import File
from cozy.ext import inject
from cozy.extensions.set import split_strings_to_set
from cozy.model.book import Book, BookIsEmpty
from cozy.model.chapter import Chapter
from cozy.model.settings import Settings

log = logging.getLogger("ui")


class Library(EventSender):
    _db = cache = inject.attr(SqliteDatabase)
    _settings: Settings = inject.attr(Settings)

    _books: List[Book] = []
    _chapters: Set[Chapter] = set()
    _files: Set[str] = set()

    def __init__(self):
        super().__init__()

    @property
    def authors(self):
        authors = {book.author for book in self.books}
        authors = split_strings_to_set(authors)
        return authors

    @property
    def readers(self):
        readers = {book.reader for book in self.books}
        readers = split_strings_to_set(readers)
        return readers

    @property
    def books(self) -> List[Book]:
        if not self._books:
            self._load_all_books()

        return self._books

    @property
    def chapters(self) -> Set[Chapter]:
        if not self._chapters:
            self._load_all_chapters()

        return self._chapters

    @property
    def files(self) -> Set[str]:
        if not self._files:
            self._load_all_files()

        return self._files

    @property
    def last_played_book(self) -> Optional[Book]:
        if not self._settings.last_played_book:
            return None

        last_book = next((book
                          for book
                          in self.books
                          if book.id == self._settings.last_played_book.id), None)

        return last_book

    @last_played_book.setter
    def last_played_book(self, new_value: Optional[Book]):
        self._settings.last_played_book = new_value

    def invalidate(self):
        for book in self._books:
            book.destroy_listeners()

        self._books = []

        for chapter in self._chapters:
            chapter.destroy_listeners()

        self._chapters = set()
        self._files = set()

    @timing
    def rebase_path(self, old_path: str, new_path: str):
        self.emit_event_main_thread("rebase-started")

        chapter_count = len(self.chapters)
        progress = 0
        for chapter in self.chapters:
            if chapter.file.startswith(old_path):
                progress += 1
                chapter.file = chapter.file.replace(old_path, new_path)
                self.emit_event_main_thread("rebase-progress", progress / chapter_count)

        self.emit_event_main_thread("rebase-finished")

    @staticmethod
    def reset_modified_date_for_all():
        File.update(modified=0).execute()

    def _load_all_books(self):
        for book_db_obj in BookModel.select():
            try:
                book = Book(self._db, book_db_obj)
                book.add_listener(self._on_book_event)
                self._books.append(book)
            except BookIsEmpty:
                pass

    def _load_all_chapters(self):
        self._chapters = {chapter
                          for book_chapters
                          in [book.chapters for book in self.books]
                          for chapter
                          in book_chapters}

        for chapter in self._chapters:
            chapter.add_listener(self._on_chapter_event)

    def _load_all_files(self):
        self._files = {chapter.file
                       for chapter
                       in self.chapters}

    def _on_chapter_event(self, event: str, chapter: Chapter):
        if event == "chapter-deleted":
            try:
                self.chapters.remove(chapter)
            except KeyError:
                log.error("Could not remove chapter from library chapter list.")

            try:
                self.files.remove(chapter.file)
            except KeyError:
                log.error("Could not remove file from library file list.")
                self._files = []

    def _on_book_event(self, event: str, book):
        if event == "book-deleted":
            self.books.remove(book)



from cozy.db.model_base import ModelBase
from peewee import CharField, IntegerField, BlobField, FloatField, BooleanField


class Book(ModelBase):
    name = CharField()
    author = CharField()
    reader = CharField()
    position = IntegerField()
    rating = IntegerField()
    cover = BlobField(null=True)
    playback_speed = FloatField(default=1.0)
    last_played = IntegerField(default=0)
    offline = BooleanField(default=False)
    downloaded = BooleanField(default=False)



import webbrowser

import gi

from cozy.ext import inject

gi.require_version('Gtk', '3.0')
from gi.repository import Gtk


@Gtk.Template(resource_path='/com/github/geigi/cozy/db_migration_failed.ui')
class DBMigrationFailedView(Gtk.Dialog):
    __gtype_name__ = 'DBMigrationFailedDialog'

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def show(self):
        response = self.run()

        if response == Gtk.ResponseType.OK:
            webbrowser.open("https://github.com/geigi/cozy/issues", new=2)



import gi

gi.require_version('Gtk', '3.0')
from gi.repository import Gtk


@Gtk.Template.from_resource('/com/github/geigi/cozy/progress_popover.ui')
class ProgressPopover(Gtk.Popover):
    __gtype_name__ = 'ProgressPopover'

    progress_label: Gtk.Label = Gtk.Template.Child()
    progress_bar: Gtk.ProgressBar = Gtk.Template.Child()

    def __init__(self):
        super().__init__()

    def set_message(self, message: str):
        self.progress_label.set_text(message)

    def set_progress(self, progress: float):
        self.progress_bar.set_fraction(progress)



import os

from peewee import SqliteDatabase

from cozy.db.storage import Storage as StorageModel


class InvalidPath(Exception):
    pass


class Storage:
    def __init__(self, db: SqliteDatabase, db_id: int):
        self._db: SqliteDatabase = db
        self.id: int = db_id

        self._get_db_object()

    def _get_db_object(self):
        self._db_object: StorageModel = StorageModel.get(self.id)

    @property
    def db_object(self):
        return self._db_object

    @property
    def path(self):
        return self._db_object.path

    @path.setter
    def path(self, new_path: str):
        if not os.path.isabs(new_path):
            raise InvalidPath

        self._db_object.path = new_path
        self._db_object.save(only=self._db_object.dirty_fields)

    @property
    def location_type(self):
        return self._db_object.location_type

    @location_type.setter
    def location_type(self, new_location_type: int):
        self._db_object.location_type = new_location_type
        self._db_object.save(only=self._db_object.dirty_fields)

    @property
    def default(self):
        return self._db_object.default

    @default.setter
    def default(self, new_default: bool):
        self._db_object.default = new_default
        self._db_object.save(only=self._db_object.dirty_fields)

    @property
    def external(self):
        return self._db_object.external

    @external.setter
    def external(self, new_external: bool):
        self._db_object.external = new_external
        self._db_object.save(only=self._db_object.dirty_fields)



from typing import List

from cozy.media.chapter import Chapter


class MediaFile:
    book_name: str
    author: str
    reader: str
    disk: int
    cover: bytes
    path: str
    modified: int
    chapters: List[Chapter]

    def __init__(self, book_name: str, author: str, reader: str, disk: int, cover: bytes, path: str, modified: int,
                 chapters: List[Chapter]):
        self.book_name = book_name
        self.author = author
        self.reader = reader
        self.disk = disk
        self.cover = cover
        self.path = path
        self.modified = modified
        self.chapters = chapters



from gi.repository import Gtk


def remove_all_children(self):
    """
    Removes all widgets from a gtk container.
    """
    self.set_visible(False)
    childs = self.get_children()
    for element in childs:
        self.remove(element)
        element.destroy()
    self.set_visible(True)


def extend_gtk_container():
    Gtk.Container.remove_all_children = remove_all_children



from cozy.architecture.event_sender import EventSender
from cozy.architecture.observable import Observable
from cozy.db.book import Book
from cozy.ext import inject
from cozy.media.player import Player


class PlaybackSpeedViewModel(Observable, EventSender):
    _player: Player = inject.attr(Player)

    def __init__(self):
        super().__init__()
        super(Observable, self).__init__()

        self._book: Book = self._player.loaded_book

        self._player.add_listener(self._on_player_event)

    @property
    def playback_speed(self) -> float:
        if self._book:
            return self._book.playback_speed
        else:
            return 1.0

    @playback_speed.setter
    def playback_speed(self, new_value: float):
        if self._book:
            self._book.playback_speed = new_value
            self._player.playback_speed = new_value

    def _on_player_event(self, event: str, message):
        if event == "chapter-changed" and message:
            self._book = message
            self._notify("playback_speed")



from gettext import gettext as _

import gi

from cozy.application_settings import ApplicationSettings
from cozy.ext import inject

gi.require_version('Gtk', '3.0')
from gi.repository import Gtk

LEVELS = [
    _("Disabled"),
    _("Basic error reporting"),
    _("Detailed error reporting"),
    _("Detailed error reporting with import errors")
]

LEVEL_DESCRIPTION = [
    _("No error or crash reporting."),
    _("The following information will be sent in case of an error or crash:")
]

LEVEL_DETAILS = {
    0: [],
    1: [_("Which type of error occurred"),
        _("Line of code where an error occurred"),
        _("Cozy's version"), ],
    2: [_("Linux distribution"),
        _("Desktop environment")],
    3: [_("Media type of files that Cozy couldn't import")]
}


@Gtk.Template.from_resource('/com/github/geigi/cozy/error_reporting.ui')
class ErrorReporting(Gtk.Box):
    __gtype_name__ = 'ErrorReporting'

    level_label: Gtk.Label = Gtk.Template.Child()
    description_label: Gtk.Label = Gtk.Template.Child()
    details_label: Gtk.Label = Gtk.Template.Child()

    verbose_adjustment: Gtk.Adjustment = Gtk.Template.Child()
    verbose_scale: Gtk.Scale = Gtk.Template.Child()

    app_settings: ApplicationSettings = inject.attr(ApplicationSettings)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.__init_scale()
        self.__connect()

        level = self.app_settings.report_level
        self.verbose_adjustment.set_value(level + 1)
        self._adjustment_changed(self.verbose_adjustment)

    def __init_scale(self):
        for i in range(1, 5):
            self.verbose_scale.add_mark(i, Gtk.PositionType.RIGHT, None)
        self.verbose_scale.set_round_digits(0)

    def __connect(self):
        self.verbose_adjustment.connect("value-changed", self._adjustment_changed)

    def _adjustment_changed(self, adjustment: Gtk.Adjustment):
        level = int(adjustment.get_value()) - 1
        self.app_settings.report_level = level
        self._update_ui_texts(level)

    def _update_ui_texts(self, level: int):
        self.level_label.set_text(LEVELS[level])
        self._update_description(level)
        self._update_details(level)

    def _update_description(self, value):
        detail_index = min(value, 1)
        self.description_label.set_text(LEVEL_DESCRIPTION[detail_index])

    def _update_details(self, value):
        details = ""
        for i in range(value + 1):
            for line in LEVEL_DETAILS[i]:
                details += "• {}\n".format(line)
        self.details_label.set_text(details)



from typing import Optional

from cozy import tools
from cozy.application_settings import ApplicationSettings
from cozy.architecture.event_sender import EventSender
from cozy.architecture.observable import Observable
from cozy.control.filesystem_monitor import FilesystemMonitor
from cozy.control.offline_cache import OfflineCache
from cozy.ext import inject
from cozy.media.player import Player
from cozy.model.book import Book
from cozy.model.chapter import Chapter
from cozy.model.library import Library
from cozy.model.settings import Settings
from cozy.open_view import OpenView


class BookDetailViewModel(Observable, EventSender):
    _player: Player = inject.attr(Player)
    _fs_monitor: FilesystemMonitor = inject.attr("FilesystemMonitor")
    _offline_cache: OfflineCache = inject.attr(OfflineCache)
    _settings: Settings = inject.attr(Settings)
    _library = Library = inject.attr(Library)
    _app_settings: ApplicationSettings = inject.attr(ApplicationSettings)

    def __init__(self):
        super().__init__()
        super(Observable, self).__init__()

        self._play = False
        self._current_chapter = None
        self._book: Book = None
        self._lock_ui: bool = False

        self._player.add_listener(self._on_player_event)
        self._fs_monitor.add_listener(self._on_fs_monitor_event)
        self._offline_cache.add_listener(self._on_offline_cache_event)
        self._app_settings.add_listener(self._on_app_setting_changed)

    @property
    def playing(self) -> bool:
        if not self._player.loaded_book or self._player.loaded_book != self._book:
            return False

        return self._player.playing

    @property
    def current_chapter(self) -> Optional[Chapter]:
        if not self.book:
            return None

        return self.book.current_chapter

    @property
    def book(self) -> Optional[Book]:
        return self._book

    @book.setter
    def book(self, value: Book):
        if self._book == value:
            return

        if self._book:
            self._book.remove_bind("current_chapter", self._on_book_current_chapter_changed)
            self._book.remove_bind("last_played", self._on_book_last_played_changed)
            self._book.remove_bind("duration", self._on_book_duration_changed)
            self._book.remove_bind("progress", self._on_book_progress_changed)
            self._book.remove_bind("playback_speed", self._on_playback_speed_changed)

        self._book = value
        self._current_chapter = None
        self._book.bind_to("current_chapter", self._on_book_current_chapter_changed)
        self._book.bind_to("last_played", self._on_book_last_played_changed)
        self._book.bind_to("duration", self._on_book_duration_changed)
        self._book.bind_to("progress", self._on_book_progress_changed)
        self._book.bind_to("playback_speed", self._on_playback_speed_changed)
        self._notify("book")

    @property
    def last_played_text(self) -> Optional[str]:
        if not self._book:
            return None

        return tools.past_date_to_human_readable(self._book.last_played)

    @property
    def total_text(self) -> Optional[str]:
        if not self._book:
            return None

        return tools.seconds_to_human_readable(self._book.duration / self._book.playback_speed)

    @property
    def remaining_text(self) -> Optional[str]:
        if not self._book:
            return None

        remaining = self._book.duration / self._book.playback_speed - self._book.progress / self._book.playback_speed
        return tools.seconds_to_human_readable(remaining)

    @property
    def progress_percent(self) -> Optional[float]:
        if not self._book:
            return None

        if self._book.duration < 1:
            return 1.0

        return self._book.progress / self._book.duration

    @property
    def disk_count(self) -> int:
        if not self._book:
            return 0

        return len({chapter.disk
                    for chapter
                    in self._book.chapters})

    @property
    def is_book_available(self) -> bool:
        if self._book.offline and self._book.downloaded:
            return True

        return self._fs_monitor.get_book_online(self._book)

    @property
    def is_book_external(self) -> bool:
        first_chapter_path = self._book.chapters[0].file
        return any(storage.path
                   in first_chapter_path
                   for storage
                   in self._settings.external_storage_locations)

    @property
    def lock_ui(self) -> bool:
        return self._lock_ui

    @lock_ui.setter
    def lock_ui(self, new_value: bool):
        self._lock_ui = new_value
        self._notify("lock_ui")

    def download_book(self, download: bool):
        self._book.offline = download

        if download:
            self._offline_cache.add(self._book)
        else:
            self._offline_cache.remove(self._book)

    def open_library(self):
        self.emit_event(OpenView.LIBRARY)

    def play_book(self):
        self._player.play_pause_book(self.book)

    def play_chapter(self, chapter: Chapter):
        if self._book.current_chapter != chapter:
            chapter.position = chapter.start_position
        self._player.play_pause_chapter(self._book, chapter)

    def open_book_detail_view(self):
        self._notify("open")

    def _on_player_event(self, event, message):
        if not self.book:
            return

        if event == "play" or event == "pause":
            self._notify("playing")
        elif event == "position" or event == "book-finished":
            self._notify("progress_percent")
            self._notify("remaining_text")

    def _on_fs_monitor_event(self, event, _):
        if not self._book:
            return

        if event == "storage-online":
            self._notify("is_book_available")
        elif event == "storage-offline":
            self._notify("is_book_available")

    def _on_book_current_chapter_changed(self):
        self._notify("current_chapter")

    def _on_book_last_played_changed(self):
        self._notify("last_played_text")

    def _on_book_progress_changed(self):
        self._notify("remaining_text")
        self._notify("progress_percent")

    def _on_book_duration_changed(self):
        self._notify("progress_percent")
        self._notify("remaining_text")
        self._notify("total_text")

    def _on_playback_speed_changed(self):
        self._notify("progress_percent")
        self._notify("remaining_text")
        self._notify("total_text")

    def _on_offline_cache_event(self, event, message):
        try:
            if message.id != self._book.id:
                return
        except Exception as e:
            return

        if event == "book-offline-removed":
            self._notify("downloaded")
        elif event == "book-offline":
            self._notify("downloaded")

    def _on_app_setting_changed(self, event, _):
        if event == "swap-author-reader":
            self._notify("book")

    def navigate_back(self):
        self.emit_event(OpenView.BACK)



from typing import Optional

from gi.repository import Gtk, Handy
from gi.repository.Gtk import Builder

from cozy.ext import inject
from cozy.ui.widgets.book_element import BookElement
from cozy.ui.delete_book_view import DeleteBookView
from cozy.ui.widgets.filter_list_box import FilterListBox
from cozy.view_model.library_view_model import LibraryViewModel, LibraryViewMode, LibraryPage

READER_PAGE = "reader"
AUTHOR_PAGE = "author"
RECENT_PAGE = "recent"
MAIN_BOOK_PAGE = "main"
NO_MEDIA_PAGE = "no_media"

NO_RECENT_PAGE = "no_recent"
BOOKS_PAGE = "books"


class LibraryView:
    _view_model: LibraryViewModel = inject.attr(LibraryViewModel)

    def __init__(self, builder: Builder):
        self._builder = builder
        self._connected_book_element: Optional[BookElement] = None

        self._get_ui_elements()
        self._connect_ui_elements()

        self._connect_view_model()

        self.populate_book_box()
        self.populate_author()
        self.populate_reader()

        self._on_library_view_mode_changed()

    def _get_ui_elements(self):
        self._filter_stack: Gtk.Stack = self._builder.get_object("sort_stack")
        self._main_stack: Gtk.Stack = self._builder.get_object("main_stack")
        self._book_box: Gtk.FlowBox = self._builder.get_object("book_box")
        self._filter_stack_revealer: Gtk.Revealer = self._builder.get_object("sort_stack_revealer")
        self._author_box: FilterListBox = self._builder.get_object("author_box")
        self._reader_box: FilterListBox = self._builder.get_object("reader_box")
        self._library_leaflet: Handy.Leaflet = self._builder.get_object("library_leaflet")
        self._book_stack: Gtk.Stack = self._builder.get_object("book_stack")

    def _connect_ui_elements(self):
        self._filter_stack.connect("notify::visible-child", self._on_sort_stack_changed)
        self._main_stack.connect("notify::visible-child", self._on_main_stack_changed)
        self._book_box.set_sort_func(self._view_model.display_book_sort)
        self._book_box.set_filter_func(self._view_model.display_book_filter)

        # We need to connect to row-activated because it will be emitted when the user clicks on a row
        # which is already activated. This is important for the mobile view
        # row-selected is needed because when setting the row using select_row (e.g. when searching), the row-activated
        # signal doesn't emit.
        self._author_box.connect("row-activated", self._on_filter_row_activated)
        self._reader_box.connect("row-activated", self._on_filter_row_activated)
        self._author_box.connect("row-selected", self._on_filter_row_activated)
        self._reader_box.connect("row-selected", self._on_filter_row_activated)

    def _connect_view_model(self):
        self._view_model.bind_to("library_view_mode", self._on_library_view_mode_changed)
        self._view_model.bind_to("library_page", self._on_library_page_changed)
        self._view_model.bind_to("authors", self.populate_author)
        self._view_model.bind_to("readers", self.populate_reader)
        self._view_model.bind_to("books", self.populate_book_box)
        self._view_model.bind_to("books-filter", self._book_box.invalidate_filter)
        self._view_model.bind_to("books-filter", self._book_box.invalidate_sort)
        self._view_model.bind_to("selected_filter", self._select_filter_row)
        self._view_model.bind_to("current_book_in_playback", self._current_book_in_playback)
        self._view_model.bind_to("playing", self._playing)
        self._view_model.bind_to("book-progress", self._on_book_progress_changed)

    def _on_sort_stack_changed(self, widget, _):
        page = widget.props.visible_child_name

        view_mode = None
        if page == RECENT_PAGE:
            view_mode = LibraryViewMode.CURRENT
        elif page == AUTHOR_PAGE:
            view_mode = LibraryViewMode.AUTHOR
        elif page == READER_PAGE:
            view_mode = LibraryViewMode.READER

        self._view_model.library_view_mode = view_mode

    def _on_main_stack_changed(self, widget, _):
        page = widget.props.visible_child_name

        if page != MAIN_BOOK_PAGE:
            self._view_model.library_page = LibraryPage.NONE

    def populate_book_box(self):
        self._book_box.remove_all_children()

        for book in self._view_model.books:
            book_element = BookElement(book)
            book_element.connect("play-pause-clicked", self._play_book_clicked)
            book_element.connect("open-book-overview", self._open_book_overview_clicked)
            book_element.connect("book-removed", self._on_book_removed)
            book_element.show_all()
            self._book_box.add(book_element)

        self._book_box.show_all()

    def populate_author(self):
        self._author_box.populate(self._view_model.authors)

    def populate_reader(self):
        self._reader_box.populate(self._view_model.readers)

    def _on_library_view_mode_changed(self):
        visible_child_name = None
        active_filter_box: Gtk.ListBox = None

        view_mode = self._view_model.library_view_mode
        main_view_page = MAIN_BOOK_PAGE
        books_view_page = BOOKS_PAGE

        if len(self._view_model.books) < 1:
            main_view_page = NO_MEDIA_PAGE
            visible_child_name = RECENT_PAGE
        elif view_mode == LibraryViewMode.CURRENT:
            visible_child_name = RECENT_PAGE
            if not self._view_model.is_any_book_in_progress:
                books_view_page = NO_RECENT_PAGE
        elif view_mode == LibraryViewMode.AUTHOR:
            visible_child_name = AUTHOR_PAGE
            active_filter_box = self._author_box
        elif view_mode == LibraryViewMode.READER:
            visible_child_name = READER_PAGE
            active_filter_box = self._reader_box

        # https://stackoverflow.com/questions/22178524/gtk-named-stack-childs/22182843#22182843
        self._main_stack.props.visible_child_name = main_view_page
        self._filter_stack.set_visible_child_name(visible_child_name)
        self._book_stack.set_visible_child_name(books_view_page)

        if active_filter_box:
            self._apply_selected_filter(active_filter_box.get_selected_row())

        self._invalidate_filters()

    def _on_library_page_changed(self):
        page = self._view_model.library_page

        if page == LibraryPage.FILTER:
            self._library_leaflet.set_visible_child_name("filter")
        elif page == LibraryPage.BOOKS:
            self._library_leaflet.set_visible_child_name("books")

    def _invalidate_filters(self):
        self._book_box.invalidate_filter()
        self._book_box.invalidate_sort()

    def _apply_selected_filter(self, row):
        if not row:
            return

        self._view_model.selected_filter = row.data
        self._invalidate_filters()

    def _on_filter_row_activated(self, _, row):
        self._apply_selected_filter(row)
        self._view_model.library_page = LibraryPage.BOOKS

        return True

    def _select_filter_row(self):
        if self._view_model.library_view_mode == LibraryViewMode.AUTHOR:
            self._author_box.select_row_with_content(self._view_model.selected_filter)
        elif self._view_model.library_view_mode == LibraryViewMode.READER:
            self._reader_box.select_row_with_content(self._view_model.selected_filter)

    def _play_book_clicked(self, _, book):
        self._view_model.play_book(book)

    def _open_book_overview_clicked(self, _, book):
        self._view_model.open_book_detail(book)
        self._view_model.library_page = LibraryPage.NONE
        return True

    def _on_book_removed(self, _, book):
        delete_from_library = True
        delete_files = False

        if self._view_model.book_files_exist(book):
            dialog = DeleteBookView()
            delete_from_library = delete_files = dialog.get_delete_book()
            dialog.destroy()

        if delete_files:
            self._view_model.delete_book_files(book)

        if delete_from_library:
            self._view_model.remove_book(book)

    def _current_book_in_playback(self):
        if self._connected_book_element:
            self._connected_book_element.set_playing(False)

        self._connected_book_element = next((book_element
                                             for book_element
                                             in self._book_box.get_children()
                                             if book_element.book == self._view_model.current_book_in_playback),
                                            None)

    def _playing(self):
        if self._connected_book_element:
            self._connected_book_element.set_playing(self._view_model.playing)

    def _on_book_progress_changed(self):
        if not self._connected_book_element:
            return

        self._connected_book_element.update_progress()



from cozy.model.chapter import Chapter


class SingleFileChapter(Chapter):
    @property
    def name(self):
        pass

    @property
    def number(self):
        pass

    @property
    def disk(self):
        pass

    @property
    def position(self):
        pass

    @property
    def file(self):
        pass

    @property
    def length(self):
        pass

    @property
    def modified(self):
        pass



import cozy.ext.inject as inject

from cozy.extensions.set import split_strings_to_set
from cozy.open_view import OpenView
from cozy.application_settings import ApplicationSettings
from cozy.architecture.event_sender import EventSender
from cozy.architecture.observable import Observable
from cozy.control.filesystem_monitor import FilesystemMonitor
from cozy.model.book import Book
from cozy.model.library import Library


class SearchViewModel(Observable, EventSender):
    _fs_monitor: FilesystemMonitor = inject.attr("FilesystemMonitor")
    _model: Library = inject.attr(Library)
    _application_settings: ApplicationSettings = inject.attr(ApplicationSettings)

    _search_open: bool = False

    def __init__(self):
        super().__init__()
        super(Observable, self).__init__()

    @property
    def books(self):
        return self._model.books

    @property
    def authors(self):
        is_book_online = self._fs_monitor.get_book_online
        show_offline_books = not self._application_settings.hide_offline

        authors = {
            book.author
            for book
            in self._model.books
            if is_book_online(book) or show_offline_books
        }

        return sorted(split_strings_to_set(authors))

    @property
    def readers(self):
        is_book_online = self._fs_monitor.get_book_online
        show_offline_books = not self._application_settings.hide_offline

        readers = {
            book.reader
            for book
            in self._model.books
            if is_book_online(book) or show_offline_books
        }

        return sorted(split_strings_to_set(readers))

    @property
    def search_open(self):
        return self._search_open

    @search_open.setter
    def search_open(self, value):
        self._search_open = value
        self._notify("search_open")

    def jump_to_book(self, book: Book):
        self.emit_event(OpenView.BOOK, book)
        self.search_open = False

    def jump_to_author(self, author: str):
        self.emit_event(OpenView.AUTHOR, author)
        self.search_open = False

    def jump_to_reader(self, reader: str):
        self.emit_event(OpenView.READER, reader)
        self.search_open = False



import re
from typing import Set


def split_strings_to_set(set_to_split: Set[str]):
    finished = set()
    for entry in set_to_split:
        results = re.split(",|;|/|&", entry)
        results = {
            entry.strip()
            for entry in results
        }

        finished.update(results)

    return finished



from peewee import CharField, IntegerField, BooleanField

from cozy.db.model_base import ModelBase


class Storage(ModelBase):
    path = CharField()
    location_type = IntegerField(default=0)
    default = BooleanField(default=False)
    external = BooleanField(default=False)



import logging
import os
import shutil
from datetime import datetime
from typing import List

from peewee import IntegerField, BooleanField, FloatField, ForeignKeyField, fn
from playhouse.migrate import SqliteMigrator, migrate
from playhouse.reflection import generate_models

from cozy.control.application_directories import get_cache_dir
from cozy.db.file import File
from cozy.db.model_base import get_data_dir, get_sqlite_database
from cozy.db.offline_cache import OfflineCache
from cozy.db.settings import Settings
from cozy.db.storage import Storage
from cozy.db.storage_blacklist import StorageBlackList
from cozy.db.track import Track
from cozy.db.track_to_file import TrackToFile
from cozy.report import reporter

log = logging.getLogger("db_updater")


def __update_db_1(db):
    migrator = SqliteMigrator(db)

    version = IntegerField(default=1)
    crc32 = BooleanField(default=False)

    migrate(
        migrator.add_column('settings', 'version', version),
        migrator.add_column('track', 'crc32', crc32),
    )


def __update_db_2(db):
    migrator = SqliteMigrator(db)

    playback_speed = FloatField(default=1.0)

    migrate(
        migrator.add_column('book', 'playback_speed', playback_speed),
    )

    Settings.update(version=2).execute()


def __update_db_3(db):
    current_path = Settings.get().path

    db.create_tables([Storage])
    Storage.create(path=current_path, default=True)
    Settings.update(path="NOT_USED").execute()
    Settings.update(version=3).execute()


def __update_db_4(db):
    migrator = SqliteMigrator(db)

    last_played = IntegerField(default=0)

    migrate(
        migrator.add_column('book', 'last_played', last_played),
    )

    Settings.update(version=4).execute()


def __update_db_5(db):
    db.create_tables([StorageBlackList])

    Settings.update(version=5).execute()


def __update_db_6(db):
    migrator = SqliteMigrator(db)

    db.create_tables([OfflineCache])

    external = BooleanField(default=False)
    offline = BooleanField(default=False)
    downloaded = BooleanField(default=False)

    migrate(
        migrator.add_column('storage', 'external', external),
        migrator.add_column('book', 'offline', offline),
        migrator.add_column('book', 'downloaded', downloaded)
    )

    Settings.update(version=6).execute()

    import shutil
    shutil.rmtree(get_cache_dir())


def __update_db_7(db):
    from cozy.control.artwork_cache import ArtworkCache
    artwork_cache = ArtworkCache()
    artwork_cache.delete_artwork_cache()
    Settings.update(version=7).execute()


def __update_db_8(db):
    db.execute_sql('UPDATE track SET modified=0 WHERE crc32=1')

    migrator: SqliteMigrator = SqliteMigrator(db)

    migrate(
        migrator.drop_column("track", "crc32")
    )

    Settings.update(version=8).execute()


def _update_db_9(db):
    log.info("Migrating to DB Version 9...")

    models = generate_models(db)

    migrator: SqliteMigrator = SqliteMigrator(db)

    db.create_tables([File, TrackToFile])
    db.stop()
    db.start()

    files: List[File] = []
    track_to_files: List[TrackToFile] = []
    file_id = 1

    if "file" in models["track"]._meta.sorted_field_names:
        log.info("Generating File and TrackToFile objects...")
        for track in models["track"].select():
            path = track.file

            file = next((f for f in files if f.path == path), None)

            if File.select().where(File.path == path).count() > 0:
                log.info("Path already existing in db: {}".format(path))
                file = File.select().where(File.path == path).get()
            elif not file:
                file = File(path=path, modified=track.modified, id=file_id)
                files.append(file)
                file_id += 1

            if TrackToFile.select().join(Track).where(TrackToFile.track.id == track.id).count() > 0:
                log.info("TrackToFile already existing in db: {}".format(path))
                continue

            track_to_file = TrackToFile(track=track, file=file, start_at=0)
            track_to_files.append(track_to_file)

        log.info("Inserting File and TrackToFile objects...")
        File.bulk_create(files, batch_size=300)
        TrackToFile.bulk_create(track_to_files, batch_size=300)

    field = ForeignKeyField(File, null=True, field=File.id)

    if "cached_file" not in models["offlinecache"]._meta.sorted_field_names:
        log.info("Rename in OfflineCache: file to cached_file...")
        migrate(
            migrator.rename_column("offlinecache", "file", "cached_file"),
        )

    if "original_file" not in models["offlinecache"]._meta.sorted_field_names:
        log.info("Add in OfflineCache: original_file_id...")
        migrate(
            migrator.add_column("offlinecache", "original_file_id", field)
        )

    db.stop()
    db.start()
    models = generate_models(db)

    if "file" in models["track"]._meta.sorted_field_names:
        log.info("Migrating OfflineCache...")
        for cache in models["offlinecache"].select():
            file_query = File.select().where(File.path == cache.track.file)
            if file_query.count() < 0:
                cache.delete_instance()

            file = file_query.get()
            cache.original_file = file
            cache.save(only=cache.dirty_fields)

    if "file" in models["track"]._meta.sorted_field_names:
        log.info("Drop in Track: file...")
        migrate(
            migrator.drop_column("track", "file")
        )

    if "modified" in models["track"]._meta.sorted_field_names:
        log.info("Drop in Track: modified...")
        migrate(
            migrator.drop_column("track", "modified")
        )

    if "track_id" in models["offlinecache"]._meta.sorted_field_names:
        log.info("Drop in OfflineCache: track_id...")
        migrate(
            migrator.drop_column("offlinecache", "track_id")
        )

    migrate(
        migrator.add_not_null("offlinecache", "original_file_id")
    )

    db.stop()
    db.start()

    log.info("Reset modified on all m4b files")
    File.update(modified=0).where(fn.Lower(File.path).endswith("m4b")).execute()

    db.stop()
    db.start()

    Settings.update(version=9).execute()


def _update_db_10(db):
    log.info("Migrating to DB Version 10...")

    models = generate_models(db)
    migrator: SqliteMigrator = SqliteMigrator(db)

    if "track" in models["offlinecache"]._meta.sorted_field_names:
        log.info("Drop in OfflineCache: track_id...")
        migrate(
            migrator.drop_column("offlinecache", "track_id")
        )

    db.stop()
    db.start()

    Settings.update(version=10).execute()


def update_db():
    db = get_sqlite_database()
    # First test for version 1
    try:
        next(c for c in db.get_columns("settings") if c.name == "version")
    except Exception as e:
        if len(db.get_tables()) == 0:
            data_dir = get_data_dir()
            if os.path.exists(os.path.join(data_dir, "cozy.db")):
                os.remove(os.path.join(data_dir, "cozy.db"))
                os.remove(os.path.join(data_dir, "cozy.db-shm"))
                os.remove(os.path.join(data_dir, "cozy.db-wal"))
        __update_db_1(db)

    version = Settings.get().version
    # then for version 2 and so on
    if version < 2:
        __update_db_2(db)

    if version < 3:
        __update_db_3(db)

    if version < 4:
        __update_db_4(db)

    if version < 5:
        __update_db_5(db)

    if version < 6:
        __update_db_6(db)

    if version < 7:
        __update_db_7(db)

    if version < 8:
        __update_db_8(db)

    if version < 9:
        backup_dir_name = _backup_db(db)
        try:
            _update_db_9(db)
        except Exception as e:
            log.error(e)
            reporter.exception("db_updator", e)
            db.stop()
            _restore_db(backup_dir_name)

            from cozy.ui.db_migration_failed_view import DBMigrationFailedView
            dialog = DBMigrationFailedView()
            dialog.show()
            exit(1)

    if version < 10:
        backup_dir_name = _backup_db(db)
        try:
            _update_db_10(db)
        except Exception as e:
            log.error(e)
            reporter.exception("db_updator", e)
            db.stop()
            _restore_db(backup_dir_name)

            from cozy.ui.db_migration_failed_view import DBMigrationFailedView
            dialog = DBMigrationFailedView()
            dialog.show()
            exit(1)


def _backup_db(db) -> str:
    log.info("Backing up DB...")
    db.stop()
    now = datetime.now()
    dt_string = now.strftime("%Y-%m-%d %H-%M-%S")
    backup_dir = os.path.join(get_data_dir(), dt_string)
    if os.path.exists(backup_dir):
        backup_dir += "-1"
    os.makedirs(backup_dir, exist_ok=True)

    db_path = os.path.join(get_data_dir(), "cozy.db")
    shm_path = os.path.join(get_data_dir(), "cozy.db-shm")
    wal_path = os.path.join(get_data_dir(), "cozy.db-wal")

    if os.path.exists(db_path):
        shutil.copyfile(db_path, os.path.join(backup_dir, "cozy.db"))

    if os.path.exists(shm_path):
        shutil.copyfile(shm_path, os.path.join(backup_dir, "cozy.db-shm"))

    if os.path.exists(wal_path):
        shutil.copyfile(wal_path, os.path.join(backup_dir, "cozy.db-wal"))

    db.start()

    return dt_string


def _restore_db(backup_dir_name: str):
    log.info("Restoring DB...")
    backup_dir = os.path.join(get_data_dir(), backup_dir_name)

    db_path = os.path.join(get_data_dir(), "cozy.db")
    shm_path = os.path.join(get_data_dir(), "cozy.db-shm")
    wal_path = os.path.join(get_data_dir(), "cozy.db-wal")

    db_path_backup = os.path.join(backup_dir, "cozy.db")
    shm_path_backup = os.path.join(backup_dir, "cozy.db-shm")
    wal_path_backup = os.path.join(backup_dir, "cozy.db-wal")

    if os.path.exists(db_path_backup):
        log.info("Copying db file")
        shutil.copyfile(db_path_backup, db_path)

    if os.path.exists(shm_path_backup):
        log.info("Copying shm file")
        shutil.copyfile(shm_path_backup, shm_path)

    if os.path.exists(wal_path_backup):
        log.info("Copying wal file")
        shutil.copyfile(wal_path_backup, wal_path)



from gi.repository import Gtk

class DiskElement(Gtk.Box):
    """
    This class represents a small disk number header for the book overview track list.
    """
    disc_number = None
    container = None

    def __init__(self, disc_number):
        super().__init__()
        self.container = Gtk.Box()

        self.disc_number = disc_number
        if disc_number > 1:
            self.container.set_margin_top(18)

        self.container.set_margin_bottom(3)
        self.container.set_margin_left(5)
        self.container.set_orientation(Gtk.Orientation.HORIZONTAL)
        self.container.get_style_context().add_class("dim-label")

        image = Gtk.Image.new_from_icon_name("media-optical-cd-audio-symbolic", Gtk.IconSize.SMALL_TOOLBAR)
        self.container.add(image)
        label = Gtk.Label()
        label.set_margin_left(5)
        text = _("Disc") + " " + str(disc_number)
        label.set_markup("<b>" + text + "</b>")
        self.container.add(label)
        self.add(self.container)



import logging

from gi.repository import Gtk

from cozy.ext import inject
from cozy.media.player import Player

log = logging.getLogger("power_mgr")


class PowerManager:
    _player: Player = inject.attr(Player)
    _gtk_app = inject.attr("GtkApp")

    def __init__(self):
        self._inhibit_cookie = None

        self._player.add_listener(self._on_player_changed)

    def _on_player_changed(self, event: str, data):
        if event in ["pause", "stop"]:
            if self._inhibit_cookie:
                log.info("Uninhibited standby.")
                self._gtk_app.uninhibit(self._inhibit_cookie)
                self._inhibit_cookie = None

        elif event == "play":
            if self._inhibit_cookie:
                return

            self._inhibit_cookie = self._gtk_app.inhibit(None, Gtk.ApplicationInhibitFlags.SUSPEND,
                                                         "Playback of audiobook")
            log.info("Inhibited standby.")



class Chapter:
    name: str
    position: int
    length: float
    number: int

    def __init__(self, name: str, position: int, length: float, number: int):
        self.name = name
        self.position = position
        self.number = number
        self.length = length
